"""
© 2024 HighFinesse Laser and Electronic Systems GmbH, Tübingen, Germany

HighFinesse Laser and Electronic Systems GmbH (hereinafter referred to as
“HighFinesse”) hereby grants, free of charge, anyone who downloads the Python dll
loading functions “wlmData.py” and “wlmConst.py” (hereinafter referred to as
“software”) from the HighFinesse website at https://www.highfinesse-downloads.com/download/t0849yd8uzpj or who receives it in any other way from
HighFinesse, a non-exclusive, non-transferable and non-sublicensable right, unlimited in
time and place, to use the software in any way or form, whether known or unknown, now
or in the future. This includes in particular the right to edit, modify, present, demonstrate,
reproduce, publish, use, digitize, distribute, broadcast, exhibit, make publicly accessible
and exploit the software.

These rights are granted only on condition that the following requirements are met:
• Any copy of the source code of the software must retain these licence terms in its
entirety. Furthermore, any edits, modifications and other changes in the copy
compared to software originally provided by HighFinesse must be clearly and
comprehensibly marked within the source code as “modified, deviating from
original files provided by HighFinesse”.
• Any copy of the software in binary form must reproduce these licence terms in its
entirety in the documentation and/or other materials provided with or connected
to the copy. Furthermore, any edits, modifications and other changes in the copy
compared to software originally provided by HighFinesse must be clearly and
comprehensibly marked as “modified, deviating from original files provided by
HighFinesse” within the documentation and/or other materials provided with or
connected to the copy.

The law of the Federal Republic of Germany shall apply exclusively. The provisions of the
UN Convention on Contracts for the International Sale of Goods (CISG) are excluded.

The courts of the Federal Republic of Germany shall have exclusive international
jurisdiction for all disputes arising from or in connection with these License terms. Among
the courts of the Federal Republic of Germany, the court having jurisdiction over the place
of business of HighFinesse shall have exclusive local jurisdiction for all disputes arising
from or in connection with these Licence terms.

HighFinesse would like to point out that the software is only a preliminary prototype and
not a final software product. HighFinesse is making this prototype available to the
experienced research community free of charge in order to promote research and enable
a critical discussion of the advantages, disadvantages and possible errors of the
software. The prototype is still under development at HighFinesse and has not yet
undergone any quality control. The software may therefore still contain errors, lack
essential functions or have other deficiencies. It should therefore only be used by highly
qualified software developers and should be extensively tested and checked before each
use in order to rule out errors and avoid any damage.

HighFinesse would be grateful for any information on errors and any suggestions for
improvement sent to info@highfinesse.de.
"""
#
# wlmData API function bindings generated from wlmData.h
#

import ctypes
import os

dll = None

def LoadDLL(path):
    global dll
    dll = ctypes.WinDLL(path) if os.name == 'nt' else ctypes.CDLL(path)

# ***********  Functions for general usage  ****************************
    # intptr_t Instantiate(int32_t RFC, int32_t Mode, intptr_t P1, int32_t P2)
    dll.Instantiate.argtypes = [ctypes.c_int32, ctypes.c_int32, ctypes.c_void_p, ctypes.c_int32]
    dll.Instantiate.restype = ctypes.c_void_p


    # void CallbackProc(int32_t Mode, int32_t IntVal, double DblVal)

    # void CallbackProcEx(int32_t Ver, int32_t Mode, int32_t IntVal, double DblVal, int32_t Res1)

    # int32_t WaitForWLMEvent(int32_t* Mode, int32_t* IntVal, double* DblVal)
    dll.WaitForWLMEvent.argtypes = [ctypes.POINTER(ctypes.c_int32), ctypes.POINTER(ctypes.c_int32), ctypes.POINTER(ctypes.c_double)]
    dll.WaitForWLMEvent.restype = ctypes.c_int32

    # int32_t WaitForWLMEventEx(int32_t* Ver, int32_t* Mode, int32_t* IntVal, double* DblVal, int32_t* Res1)
    dll.WaitForWLMEventEx.argtypes = [ctypes.POINTER(ctypes.c_int32), ctypes.POINTER(ctypes.c_int32), ctypes.POINTER(ctypes.c_int32), ctypes.POINTER(ctypes.c_double), ctypes.POINTER(ctypes.c_int32)]
    dll.WaitForWLMEventEx.restype = ctypes.c_int32

    # int32_t WaitForNextWLMEvent(int32_t* Mode, int32_t* IntVal, double* DblVal)
    dll.WaitForNextWLMEvent.argtypes = [ctypes.POINTER(ctypes.c_int32), ctypes.POINTER(ctypes.c_int32), ctypes.POINTER(ctypes.c_double)]
    dll.WaitForNextWLMEvent.restype = ctypes.c_int32

    # int32_t WaitForNextWLMEventEx(int32_t* Ver, int32_t* Mode, int32_t* IntVal, double* DblVal, int32_t* Res1)
    dll.WaitForNextWLMEventEx.argtypes = [ctypes.POINTER(ctypes.c_int32), ctypes.POINTER(ctypes.c_int32), ctypes.POINTER(ctypes.c_int32), ctypes.POINTER(ctypes.c_double), ctypes.POINTER(ctypes.c_int32)]
    dll.WaitForNextWLMEventEx.restype = ctypes.c_int32
    
    # void ClearWLMEvents(void)
    dll.ClearWLMEvents.argtypes = []
    dll.ClearWLMEvents.restype = None


    # int32_t ControlWLM(int32_t Action, intptr_t App, int32_t Ver)
    dll.ControlWLM.argtypes = [ctypes.c_int32, ctypes.c_void_p, ctypes.c_int32]
    dll.ControlWLM.restype = ctypes.c_int32

    # int32_t ControlWLMEx(int32_t Action, intptr_t App, int32_t Ver, int32_t Delay, int32_t Res)
    dll.ControlWLMEx.argtypes = [ctypes.c_int32, ctypes.c_void_p, ctypes.c_int32, ctypes.c_int32, ctypes.c_int32]
    dll.ControlWLMEx.restype = ctypes.c_int32

    # int64_t SynchroniseWLM(int32_t Mode, int64_t TS)
    dll.SynchroniseWLM.argtypes = [ctypes.c_int32, ctypes.c_int64]
    dll.SynchroniseWLM.restype = ctypes.c_int64

    # int32_t SetMeasurementDelayMethod(int32_t Mode, int32_t Delay)
    dll.SetMeasurementDelayMethod.argtypes = [ctypes.c_int32, ctypes.c_int32]
    dll.SetMeasurementDelayMethod.restype = ctypes.c_int32

    # int32_t SetWLMPriority(int32_t PPC, int32_t Res1, int32_t Res2)
    dll.SetWLMPriority.argtypes = [ctypes.c_int32, ctypes.c_int32, ctypes.c_int32]
    dll.SetWLMPriority.restype = ctypes.c_int32

    # int32_t PresetWLMIndex(int32_t Ver)
    dll.PresetWLMIndex.argtypes = [ctypes.c_int32]
    dll.PresetWLMIndex.restype = ctypes.c_int32


    # int32_t GetWLMVersion(int32_t Ver)
    dll.GetWLMVersion.argtypes = [ctypes.c_int32]
    dll.GetWLMVersion.restype = ctypes.c_int32

    # int32_t GetWLMIndex(int32_t Ver)
    dll.GetWLMIndex.argtypes = [ctypes.c_int32]
    dll.GetWLMIndex.restype = ctypes.c_int32

    # int32_t GetWLMCount(int32_t V)
    dll.GetWLMCount.argtypes = [ctypes.c_int32]
    dll.GetWLMCount.restype = ctypes.c_int32

    # int32_t GetOptionInfo(int32_t Index, int32_t Detail, int64_t* I64Val, double* DblVal)
    dll.GetOptionInfo.argtypes = [ctypes.c_int32, ctypes.c_int32, ctypes.POINTER(ctypes.c_int64), ctypes.POINTER(ctypes.c_double)]
    dll.GetOptionInfo.restype = ctypes.c_int32


# ***********  General Get... & Set...-functions  **********************
    # double GetWavelength(double WL)
    dll.GetWavelength.argtypes = [ctypes.c_double]
    dll.GetWavelength.restype = ctypes.c_double

    # double GetWavelength2(double WL2)
    dll.GetWavelength2.argtypes = [ctypes.c_double]
    dll.GetWavelength2.restype = ctypes.c_double

    # double GetWavelengthNum(int32_t num, double WL)
    dll.GetWavelengthNum.argtypes = [ctypes.c_int32, ctypes.c_double]
    dll.GetWavelengthNum.restype = ctypes.c_double

    # double GetCalWavelength(int32_t ba, double WL)
    dll.GetCalWavelength.argtypes = [ctypes.c_int32, ctypes.c_double]
    dll.GetCalWavelength.restype = ctypes.c_double

    # double GetCalibrationEffect(double CE)
    dll.GetCalibrationEffect.argtypes = [ctypes.c_double]
    dll.GetCalibrationEffect.restype = ctypes.c_double

    # double GetFrequency(double F)
    dll.GetFrequency.argtypes = [ctypes.c_double]
    dll.GetFrequency.restype = ctypes.c_double

    # double GetFrequency2(double F2)
    dll.GetFrequency2.argtypes = [ctypes.c_double]
    dll.GetFrequency2.restype = ctypes.c_double

    # double GetFrequencyNum(int32_t num, double F)
    dll.GetFrequencyNum.argtypes = [ctypes.c_int32, ctypes.c_double]
    dll.GetFrequencyNum.restype = ctypes.c_double

    # double GetLinewidth(int32_t Index, double LW)
    dll.GetLinewidth.argtypes = [ctypes.c_int32, ctypes.c_double]
    dll.GetLinewidth.restype = ctypes.c_double

    # double GetLinewidthNum(int32_t num, double LW)
    dll.GetLinewidthNum.argtypes = [ctypes.c_int32, ctypes.c_double]
    dll.GetLinewidthNum.restype = ctypes.c_double

    # double GetDistance(double D)
    dll.GetDistance.argtypes = [ctypes.c_double]
    dll.GetDistance.restype = ctypes.c_double

    # double GetAnalogIn(double AI)
    dll.GetAnalogIn.argtypes = [ctypes.c_double]
    dll.GetAnalogIn.restype = ctypes.c_double

    # double GetMultimodeInfo(int32_t num, int32_t type, int32_t mode, double* Val)
    dll.GetMultimodeInfo.argtypes = [ctypes.c_int32, ctypes.c_int32, ctypes.c_int32, ctypes.POINTER(ctypes.c_double)]
    dll.GetMultimodeInfo.restype = ctypes.c_double

    # double GetTemperature(double T)
    dll.GetTemperature.argtypes = [ctypes.c_double]
    dll.GetTemperature.restype = ctypes.c_double

    # int32_t SetTemperature(double T)
    dll.SetTemperature.argtypes = [ctypes.c_double]
    dll.SetTemperature.restype = ctypes.c_int32

    # double GetPressure(double P)
    dll.GetPressure.argtypes = [ctypes.c_double]
    dll.GetPressure.restype = ctypes.c_double

    # int32_t SetPressure(int32_t Mode, double P)
    dll.SetPressure.argtypes = [ctypes.c_int32, ctypes.c_double]
    dll.SetPressure.restype = ctypes.c_int32

    # int32_t GetAirParameters(int32_t Mode, int32_t* State, double* Val)
    dll.GetAirParameters.argtypes = [ctypes.c_int32, ctypes.POINTER(ctypes.c_int32), ctypes.POINTER(ctypes.c_double)]
    dll.GetAirParameters.restype = ctypes.c_int32

    # int32_t SetAirParameters(int32_t Mode, int32_t State, double Val)
    dll.SetAirParameters.argtypes = [ctypes.c_int32, ctypes.c_int32, ctypes.c_double]
    dll.SetAirParameters.restype = ctypes.c_int32

    # double GetExternalInput(int32_t Index, double I)
    dll.GetExternalInput.argtypes = [ctypes.c_int32, ctypes.c_double]
    dll.GetExternalInput.restype = ctypes.c_double

    # int32_t SetExternalInput(int32_t Index, double I)
    dll.SetExternalInput.argtypes = [ctypes.c_int32, ctypes.c_double]
    dll.SetExternalInput.restype = ctypes.c_int32

    # int32_t GetExtraSetting(int32_t Index, int32_t* lGet, double* dGet, char* sGet)
    dll.GetExtraSetting.argtypes = [ctypes.c_int32, ctypes.POINTER(ctypes.c_int32), ctypes.POINTER(ctypes.c_double), ctypes.POINTER(ctypes.c_char)]
    dll.GetExtraSetting.restype = ctypes.c_int32

    # int32_t SetExtraSetting(int32_t Index, int32_t lSet, double dSet, char* sSet)
    dll.SetExtraSetting.argtypes = [ctypes.c_int32, ctypes.c_int32, ctypes.c_double, ctypes.POINTER(ctypes.c_char)]
    dll.SetExtraSetting.restype = ctypes.c_int32


    # uint16_t GetExposure(uint16_t E)
    dll.GetExposure.argtypes = [ctypes.c_uint16]
    dll.GetExposure.restype = ctypes.c_uint16

    # int32_t SetExposure(uint16_t E)
    dll.SetExposure.argtypes = [ctypes.c_uint16]
    dll.SetExposure.restype = ctypes.c_int32

    # uint16_t GetExposure2(uint16_t E2)
    dll.GetExposure2.argtypes = [ctypes.c_uint16]
    dll.GetExposure2.restype = ctypes.c_uint16

    # int32_t SetExposure2(uint16_t E2)
    dll.SetExposure2.argtypes = [ctypes.c_uint16]
    dll.SetExposure2.restype = ctypes.c_int32

    # int32_t GetExposureNum(int32_t num, int32_t arr, int32_t E)
    dll.GetExposureNum.argtypes = [ctypes.c_int32, ctypes.c_int32, ctypes.c_int32]
    dll.GetExposureNum.restype = ctypes.c_int32

    # int32_t SetExposureNum(int32_t num, int32_t arr, int32_t E)
    dll.SetExposureNum.argtypes = [ctypes.c_int32, ctypes.c_int32, ctypes.c_int32]
    dll.SetExposureNum.restype = ctypes.c_int32

    # double GetExposureNumEx(int32_t num, int32_t arr, double E)
    dll.GetExposureNumEx.argtypes = [ctypes.c_int32, ctypes.c_int32, ctypes.c_double]
    dll.GetExposureNumEx.restype = ctypes.c_double

    # int32_t SetExposureNumEx(int32_t num, int32_t arr, double E)
    dll.SetExposureNumEx.argtypes = [ctypes.c_int32, ctypes.c_int32, ctypes.c_double]
    dll.SetExposureNumEx.restype = ctypes.c_int32

    # bool GetExposureMode(bool EM)
    dll.GetExposureMode.argtypes = [ctypes.c_bool]
    dll.GetExposureMode.restype = ctypes.c_bool

    # int32_t SetExposureMode(bool EM)
    dll.SetExposureMode.argtypes = [ctypes.c_bool]
    dll.SetExposureMode.restype = ctypes.c_int32

    # int32_t GetExposureModeNum(int32_t num, bool EM)
    dll.GetExposureModeNum.argtypes = [ctypes.c_int32, ctypes.c_bool]
    dll.GetExposureModeNum.restype = ctypes.c_int32

    # int32_t SetExposureModeNum(int32_t num, bool EM)
    dll.SetExposureModeNum.argtypes = [ctypes.c_int32, ctypes.c_bool]
    dll.SetExposureModeNum.restype = ctypes.c_int32

    # int32_t GetExposureRange(int32_t ER)
    dll.GetExposureRange.argtypes = [ctypes.c_int32]
    dll.GetExposureRange.restype = ctypes.c_int32

    # double GetExposureRangeEx(int32_t ER)
    dll.GetExposureRangeEx.argtypes = [ctypes.c_int32]
    dll.GetExposureRangeEx.restype = ctypes.c_double

    # int32_t GetAutoExposureSetting(int32_t num, int32_t AES, int32_t* iVal, double* dVal)
    dll.GetAutoExposureSetting.argtypes = [ctypes.c_int32, ctypes.c_int32, ctypes.POINTER(ctypes.c_int32), ctypes.POINTER(ctypes.c_double)]
    dll.GetAutoExposureSetting.restype = ctypes.c_int32

    # int32_t SetAutoExposureSetting(int32_t num, int32_t AES, int32_t iVal, double dVal)
    dll.SetAutoExposureSetting.argtypes = [ctypes.c_int32, ctypes.c_int32, ctypes.c_int32, ctypes.c_double]
    dll.SetAutoExposureSetting.restype = ctypes.c_int32


    # uint16_t GetResultMode(uint16_t RM)
    dll.GetResultMode.argtypes = [ctypes.c_uint16]
    dll.GetResultMode.restype = ctypes.c_uint16

    # int32_t SetResultMode(uint16_t RM)
    dll.SetResultMode.argtypes = [ctypes.c_uint16]
    dll.SetResultMode.restype = ctypes.c_int32

    # uint16_t GetRange(uint16_t R)
    dll.GetRange.argtypes = [ctypes.c_uint16]
    dll.GetRange.restype = ctypes.c_uint16

    # int32_t SetRange(uint16_t R)
    dll.SetRange.argtypes = [ctypes.c_uint16]
    dll.SetRange.restype = ctypes.c_int32

    # uint16_t GetPulseMode(uint16_t PM)
    dll.GetPulseMode.argtypes = [ctypes.c_uint16]
    dll.GetPulseMode.restype = ctypes.c_uint16

    # int32_t SetPulseMode(uint16_t PM)
    dll.SetPulseMode.argtypes = [ctypes.c_uint16]
    dll.SetPulseMode.restype = ctypes.c_int32

    # int32_t GetPulseDelay(int32_t PD)
    dll.GetPulseDelay.argtypes = [ctypes.c_int32]
    dll.GetPulseDelay.restype = ctypes.c_int32

    # int32_t SetPulseDelay(int32_t PD)
    dll.SetPulseDelay.argtypes = [ctypes.c_int32]
    dll.SetPulseDelay.restype = ctypes.c_int32

    # uint16_t GetWideMode(uint16_t WM)
    dll.GetWideMode.argtypes = [ctypes.c_uint16]
    dll.GetWideMode.restype = ctypes.c_uint16

    # int32_t SetWideMode(uint16_t WM)
    dll.SetWideMode.argtypes = [ctypes.c_uint16]
    dll.SetWideMode.restype = ctypes.c_int32


    # int32_t GetDisplayMode(int32_t DM)
    dll.GetDisplayMode.argtypes = [ctypes.c_int32]
    dll.GetDisplayMode.restype = ctypes.c_int32

    # int32_t SetDisplayMode(int32_t DM)
    dll.SetDisplayMode.argtypes = [ctypes.c_int32]
    dll.SetDisplayMode.restype = ctypes.c_int32

    # bool GetFastMode(bool FM)
    dll.GetFastMode.argtypes = [ctypes.c_bool]
    dll.GetFastMode.restype = ctypes.c_bool

    # int32_t SetFastMode(bool FM)
    dll.SetFastMode.argtypes = [ctypes.c_bool]
    dll.SetFastMode.restype = ctypes.c_int32


    # bool GetLinewidthMode(bool LM)
    dll.GetLinewidthMode.argtypes = [ctypes.c_bool]
    dll.GetLinewidthMode.restype = ctypes.c_bool

    # int32_t SetLinewidthMode(bool LM)
    dll.SetLinewidthMode.argtypes = [ctypes.c_bool]
    dll.SetLinewidthMode.restype = ctypes.c_int32


    # bool GetDistanceMode(bool DM)
    dll.GetDistanceMode.argtypes = [ctypes.c_bool]
    dll.GetDistanceMode.restype = ctypes.c_bool

    # int32_t SetDistanceMode(bool DM)
    dll.SetDistanceMode.argtypes = [ctypes.c_bool]
    dll.SetDistanceMode.restype = ctypes.c_int32


    # int32_t GetSwitcherMode(int32_t SM)
    dll.GetSwitcherMode.argtypes = [ctypes.c_int32]
    dll.GetSwitcherMode.restype = ctypes.c_int32

    # int32_t SetSwitcherMode(int32_t SM)
    dll.SetSwitcherMode.argtypes = [ctypes.c_int32]
    dll.SetSwitcherMode.restype = ctypes.c_int32

    # int32_t GetSwitcherChannel(int32_t CH)
    dll.GetSwitcherChannel.argtypes = [ctypes.c_int32]
    dll.GetSwitcherChannel.restype = ctypes.c_int32

    # int32_t SetSwitcherChannel(int32_t CH)
    dll.SetSwitcherChannel.argtypes = [ctypes.c_int32]
    dll.SetSwitcherChannel.restype = ctypes.c_int32

    # int32_t GetSwitcherSignalStates(int32_t Signal, int32_t* Use, int32_t* Show)
    dll.GetSwitcherSignalStates.argtypes = [ctypes.c_int32, ctypes.POINTER(ctypes.c_int32), ctypes.POINTER(ctypes.c_int32)]
    dll.GetSwitcherSignalStates.restype = ctypes.c_int32

    # int32_t SetSwitcherSignalStates(int32_t Signal, int32_t Use, int32_t Show)
    dll.SetSwitcherSignalStates.argtypes = [ctypes.c_int32, ctypes.c_int32, ctypes.c_int32]
    dll.SetSwitcherSignalStates.restype = ctypes.c_int32

    # int32_t SetSwitcherSignal(int32_t Signal, int32_t Use, int32_t Show)
    dll.SetSwitcherSignal.argtypes = [ctypes.c_int32, ctypes.c_int32, ctypes.c_int32]
    dll.SetSwitcherSignal.restype = ctypes.c_int32


    # int32_t GetAutoCalMode(int32_t ACM)
    dll.GetAutoCalMode.argtypes = [ctypes.c_int32]
    dll.GetAutoCalMode.restype = ctypes.c_int32

    # int32_t SetAutoCalMode(int32_t ACM)
    dll.SetAutoCalMode.argtypes = [ctypes.c_int32]
    dll.SetAutoCalMode.restype = ctypes.c_int32

    # int32_t GetAutoCalSetting(int32_t ACS, int32_t* val, int32_t Res1, int32_t* Res2)
    dll.GetAutoCalSetting.argtypes = [ctypes.c_int32, ctypes.POINTER(ctypes.c_int32), ctypes.c_int32, ctypes.POINTER(ctypes.c_int32)]
    dll.GetAutoCalSetting.restype = ctypes.c_int32

    # int32_t SetAutoCalSetting(int32_t ACS, int32_t val, int32_t Res1, int32_t Res2)
    dll.SetAutoCalSetting.argtypes = [ctypes.c_int32, ctypes.c_int32, ctypes.c_int32, ctypes.c_int32]
    dll.SetAutoCalSetting.restype = ctypes.c_int32


    # int32_t GetActiveChannel(int32_t Mode, int32_t* Port, int32_t Res1)
    dll.GetActiveChannel.argtypes = [ctypes.c_int32, ctypes.POINTER(ctypes.c_int32), ctypes.c_int32]
    dll.GetActiveChannel.restype = ctypes.c_int32

    # int32_t SetActiveChannel(int32_t Mode, int32_t Port, int32_t CH, int32_t Res1)
    dll.SetActiveChannel.argtypes = [ctypes.c_int32, ctypes.c_int32, ctypes.c_int32, ctypes.c_int32]
    dll.SetActiveChannel.restype = ctypes.c_int32

    # int32_t GetChannelsCount(int32_t C)
    dll.GetChannelsCount.argtypes = [ctypes.c_int32]
    dll.GetChannelsCount.restype = ctypes.c_int32


    # uint16_t GetOperationState(uint16_t OS)
    dll.GetOperationState.argtypes = [ctypes.c_uint16]
    dll.GetOperationState.restype = ctypes.c_uint16

    # int32_t Operation(uint16_t Op)
    dll.Operation.argtypes = [ctypes.c_uint16]
    dll.Operation.restype = ctypes.c_int32

    # int32_t SetOperationFile(char* lpFile)
    dll.SetOperationFile.argtypes = [ctypes.POINTER(ctypes.c_char)]
    dll.SetOperationFile.restype = ctypes.c_int32

    # int32_t Calibration(int32_t Type, int32_t Unit, double Value, int32_t Channel)
    dll.Calibration.argtypes = [ctypes.c_int32, ctypes.c_int32, ctypes.c_double, ctypes.c_int32]
    dll.Calibration.restype = ctypes.c_int32

    # int32_t RaiseMeasurementEvent(int32_t Mode)
    dll.RaiseMeasurementEvent.argtypes = [ctypes.c_int32]
    dll.RaiseMeasurementEvent.restype = ctypes.c_int32

    # int32_t TriggerMeasurement(int32_t Action)
    dll.TriggerMeasurement.argtypes = [ctypes.c_int32]
    dll.TriggerMeasurement.restype = ctypes.c_int32

    # int32_t GetTriggerState(int32_t TS)
    dll.GetTriggerState.argtypes = [ctypes.c_int32]
    dll.GetTriggerState.restype = ctypes.c_int32

    # int32_t GetInterval(int32_t I)
    dll.GetInterval.argtypes = [ctypes.c_int32]
    dll.GetInterval.restype = ctypes.c_int32

    # int32_t SetInterval(int32_t I)
    dll.SetInterval.argtypes = [ctypes.c_int32]
    dll.SetInterval.restype = ctypes.c_int32

    # bool GetIntervalMode(bool IM)
    dll.GetIntervalMode.argtypes = [ctypes.c_bool]
    dll.GetIntervalMode.restype = ctypes.c_bool

    # int32_t SetIntervalMode(bool IM)
    dll.SetIntervalMode.argtypes = [ctypes.c_bool]
    dll.SetIntervalMode.restype = ctypes.c_int32

    # double GetInternalTriggerRate(double TR)
    dll.GetInternalTriggerRate.argtypes = [ctypes.c_double]
    dll.GetInternalTriggerRate.restype = ctypes.c_double

    # int32_t SetInternalTriggerRate(double TR)
    dll.SetInternalTriggerRate.argtypes = [ctypes.c_double]
    dll.SetInternalTriggerRate.restype = ctypes.c_int32

    # int32_t GetBackground(int32_t BG)
    dll.GetBackground.argtypes = [ctypes.c_int32]
    dll.GetBackground.restype = ctypes.c_int32

    # int32_t SetBackground(int32_t BG)
    dll.SetBackground.argtypes = [ctypes.c_int32]
    dll.SetBackground.restype = ctypes.c_int32

    # int32_t GetAveragingSettingNum(int32_t num, int32_t AS, int32_t Value)
    dll.GetAveragingSettingNum.argtypes = [ctypes.c_int32, ctypes.c_int32, ctypes.c_int32]
    dll.GetAveragingSettingNum.restype = ctypes.c_int32

    # int32_t SetAveragingSettingNum(int32_t num, int32_t AS, int32_t Value)
    dll.SetAveragingSettingNum.argtypes = [ctypes.c_int32, ctypes.c_int32, ctypes.c_int32]
    dll.SetAveragingSettingNum.restype = ctypes.c_int32


    # bool GetLinkState(bool LS)
    dll.GetLinkState.argtypes = [ctypes.c_bool]
    dll.GetLinkState.restype = ctypes.c_bool

    # int32_t SetLinkState(bool LS)
    dll.SetLinkState.argtypes = [ctypes.c_bool]
    dll.SetLinkState.restype = ctypes.c_int32

    # void LinkSettingsDlg(void)
    dll.LinkSettingsDlg.argtypes = []
    dll.LinkSettingsDlg.restype = None


    # int32_t GetPatternItemSize(int32_t Index)
    dll.GetPatternItemSize.argtypes = [ctypes.c_int32]
    dll.GetPatternItemSize.restype = ctypes.c_int32

    # int32_t GetPatternItemCount(int32_t Index)
    dll.GetPatternItemCount.argtypes = [ctypes.c_int32]
    dll.GetPatternItemCount.restype = ctypes.c_int32

    # intptr_t GetPattern(int32_t Index)
    dll.GetPattern.argtypes = [ctypes.c_int32]
    dll.GetPattern.restype = ctypes.c_void_p

    # intptr_t GetPatternNum(int32_t Chn, int32_t Index)
    dll.GetPatternNum.argtypes = [ctypes.c_int32, ctypes.c_int32]
    dll.GetPatternNum.restype = ctypes.c_void_p

    # int32_t GetPatternData(int32_t Index, intptr_t PArray)
    dll.GetPatternData.argtypes = [ctypes.c_int32, ctypes.c_void_p]
    dll.GetPatternData.restype = ctypes.c_int32

    # int32_t GetPatternDataNum(int32_t Chn, int32_t Index, intptr_t PArray)
    dll.GetPatternDataNum.argtypes = [ctypes.c_int32, ctypes.c_int32, ctypes.c_void_p]
    dll.GetPatternDataNum.restype = ctypes.c_int32

    # int32_t SetPattern(int32_t Index, int32_t iEnable)
    dll.SetPattern.argtypes = [ctypes.c_int32, ctypes.c_int32]
    dll.SetPattern.restype = ctypes.c_int32

    # int32_t SetPatternData(int32_t Index, intptr_t PArray)
    dll.SetPatternData.argtypes = [ctypes.c_int32, ctypes.c_void_p]
    dll.SetPatternData.restype = ctypes.c_int32


    # bool GetAnalysisMode(bool AM)
    dll.GetAnalysisMode.argtypes = [ctypes.c_bool]
    dll.GetAnalysisMode.restype = ctypes.c_bool

    # int32_t SetAnalysisMode(bool AM)
    dll.SetAnalysisMode.argtypes = [ctypes.c_bool]
    dll.SetAnalysisMode.restype = ctypes.c_int32

    # int32_t GetAnalysisItemSize(int32_t Index)
    dll.GetAnalysisItemSize.argtypes = [ctypes.c_int32]
    dll.GetAnalysisItemSize.restype = ctypes.c_int32

    # int32_t GetAnalysisItemCount(int32_t Index)
    dll.GetAnalysisItemCount.argtypes = [ctypes.c_int32]
    dll.GetAnalysisItemCount.restype = ctypes.c_int32

    # intptr_t GetAnalysis(int32_t Index)
    dll.GetAnalysis.argtypes = [ctypes.c_int32]
    dll.GetAnalysis.restype = ctypes.c_void_p

    # int32_t GetAnalysisData(int32_t Index, intptr_t PArray)
    dll.GetAnalysisData.argtypes = [ctypes.c_int32, ctypes.c_void_p]
    dll.GetAnalysisData.restype = ctypes.c_int32

    # int32_t SetAnalysis(int32_t Index, int32_t iEnable)
    dll.SetAnalysis.argtypes = [ctypes.c_int32, ctypes.c_int32]
    dll.SetAnalysis.restype = ctypes.c_int32


    # int32_t GetMinPeak(int32_t M1)
    dll.GetMinPeak.argtypes = [ctypes.c_int32]
    dll.GetMinPeak.restype = ctypes.c_int32

    # int32_t GetMinPeak2(int32_t M2)
    dll.GetMinPeak2.argtypes = [ctypes.c_int32]
    dll.GetMinPeak2.restype = ctypes.c_int32

    # int32_t GetMaxPeak(int32_t X1)
    dll.GetMaxPeak.argtypes = [ctypes.c_int32]
    dll.GetMaxPeak.restype = ctypes.c_int32

    # int32_t GetMaxPeak2(int32_t X2)
    dll.GetMaxPeak2.argtypes = [ctypes.c_int32]
    dll.GetMaxPeak2.restype = ctypes.c_int32

    # int32_t GetAvgPeak(int32_t A1)
    dll.GetAvgPeak.argtypes = [ctypes.c_int32]
    dll.GetAvgPeak.restype = ctypes.c_int32

    # int32_t GetAvgPeak2(int32_t A2)
    dll.GetAvgPeak2.argtypes = [ctypes.c_int32]
    dll.GetAvgPeak2.restype = ctypes.c_int32

    # int32_t SetAvgPeak(int32_t PA)
    dll.SetAvgPeak.argtypes = [ctypes.c_int32]
    dll.SetAvgPeak.restype = ctypes.c_int32


    # int32_t GetAmplitudeNum(int32_t num, int32_t Index, int32_t A)
    dll.GetAmplitudeNum.argtypes = [ctypes.c_int32, ctypes.c_int32, ctypes.c_int32]
    dll.GetAmplitudeNum.restype = ctypes.c_int32

    # double GetIntensityNum(int32_t num, double I)
    dll.GetIntensityNum.argtypes = [ctypes.c_int32, ctypes.c_double]
    dll.GetIntensityNum.restype = ctypes.c_double

    # double GetPowerNum(int32_t num, double P)
    dll.GetPowerNum.argtypes = [ctypes.c_int32, ctypes.c_double]
    dll.GetPowerNum.restype = ctypes.c_double


    # uint16_t GetDelay(uint16_t D)
    dll.GetDelay.argtypes = [ctypes.c_uint16]
    dll.GetDelay.restype = ctypes.c_uint16

    # int32_t SetDelay(uint16_t D)
    dll.SetDelay.argtypes = [ctypes.c_uint16]
    dll.SetDelay.restype = ctypes.c_int32

    # uint16_t GetShift(uint16_t S)
    dll.GetShift.argtypes = [ctypes.c_uint16]
    dll.GetShift.restype = ctypes.c_uint16

    # int32_t SetShift(uint16_t S)
    dll.SetShift.argtypes = [ctypes.c_uint16]
    dll.SetShift.restype = ctypes.c_int32

    # uint16_t GetShift2(uint16_t S2)
    dll.GetShift2.argtypes = [ctypes.c_uint16]
    dll.GetShift2.restype = ctypes.c_uint16

    # int32_t SetShift2(uint16_t S2)
    dll.SetShift2.argtypes = [ctypes.c_uint16]
    dll.SetShift2.restype = ctypes.c_int32

    # double GetGain(int32_t num, int32_t index, int32_t mode, double* Gain)
    dll.GetGain.argtypes = [ctypes.c_int32, ctypes.c_int32, ctypes.c_int32, ctypes.POINTER(ctypes.c_double)]
    dll.GetGain.restype = ctypes.c_double

    # int32_t SetGain(int32_t num, int32_t index, int32_t mode, double Gain)
    dll.SetGain.argtypes = [ctypes.c_int32, ctypes.c_int32, ctypes.c_int32, ctypes.c_double]
    dll.SetGain.restype = ctypes.c_int32


# ***********  Deviation (Laser Control) and PID-functions  ************
    # bool GetDeviationMode(bool DM)
    dll.GetDeviationMode.argtypes = [ctypes.c_bool]
    dll.GetDeviationMode.restype = ctypes.c_bool

    # int32_t SetDeviationMode(bool DM)
    dll.SetDeviationMode.argtypes = [ctypes.c_bool]
    dll.SetDeviationMode.restype = ctypes.c_int32

    # double GetDeviationReference(double DR)
    dll.GetDeviationReference.argtypes = [ctypes.c_double]
    dll.GetDeviationReference.restype = ctypes.c_double

    # int32_t SetDeviationReference(double DR)
    dll.SetDeviationReference.argtypes = [ctypes.c_double]
    dll.SetDeviationReference.restype = ctypes.c_int32

    # int32_t GetDeviationSensitivity(int32_t DS)
    dll.GetDeviationSensitivity.argtypes = [ctypes.c_int32]
    dll.GetDeviationSensitivity.restype = ctypes.c_int32

    # int32_t SetDeviationSensitivity(int32_t DS)
    dll.SetDeviationSensitivity.argtypes = [ctypes.c_int32]
    dll.SetDeviationSensitivity.restype = ctypes.c_int32

    # double GetDeviationSignal(double DS)
    dll.GetDeviationSignal.argtypes = [ctypes.c_double]
    dll.GetDeviationSignal.restype = ctypes.c_double

    # double GetDeviationSignalNum(int32_t Port, double DS)
    dll.GetDeviationSignalNum.argtypes = [ctypes.c_int32, ctypes.c_double]
    dll.GetDeviationSignalNum.restype = ctypes.c_double

    # int32_t SetDeviationSignal(double DS)
    dll.SetDeviationSignal.argtypes = [ctypes.c_double]
    dll.SetDeviationSignal.restype = ctypes.c_int32

    # int32_t SetDeviationSignalNum(int32_t Port, double DS)
    dll.SetDeviationSignalNum.argtypes = [ctypes.c_int32, ctypes.c_double]
    dll.SetDeviationSignalNum.restype = ctypes.c_int32

    # double RaiseDeviationSignal(int32_t iType, double dSignal)
    dll.RaiseDeviationSignal.argtypes = [ctypes.c_int32, ctypes.c_double]
    dll.RaiseDeviationSignal.restype = ctypes.c_double


    # int32_t GetPIDCourse(char* PIDC)
    dll.GetPIDCourse.argtypes = [ctypes.POINTER(ctypes.c_char)]
    dll.GetPIDCourse.restype = ctypes.c_int32

    # int32_t SetPIDCourse(char* PIDC)
    dll.SetPIDCourse.argtypes = [ctypes.POINTER(ctypes.c_char)]
    dll.SetPIDCourse.restype = ctypes.c_int32

    # int32_t GetPIDCourseNum(int32_t Port, char* PIDC)
    dll.GetPIDCourseNum.argtypes = [ctypes.c_int32, ctypes.POINTER(ctypes.c_char)]
    dll.GetPIDCourseNum.restype = ctypes.c_int32

    # int32_t SetPIDCourseNum(int32_t Port, char* PIDC)
    dll.SetPIDCourseNum.argtypes = [ctypes.c_int32, ctypes.POINTER(ctypes.c_char)]
    dll.SetPIDCourseNum.restype = ctypes.c_int32

    # int32_t GetPIDSetting(int32_t PS, int32_t Port, int32_t* iSet, double* dSet)
    dll.GetPIDSetting.argtypes = [ctypes.c_int32, ctypes.c_int32, ctypes.POINTER(ctypes.c_int32), ctypes.POINTER(ctypes.c_double)]
    dll.GetPIDSetting.restype = ctypes.c_int32

    # int32_t SetPIDSetting(int32_t PS, int32_t Port, int32_t iSet, double dSet)
    dll.SetPIDSetting.argtypes = [ctypes.c_int32, ctypes.c_int32, ctypes.c_int32, ctypes.c_double]
    dll.SetPIDSetting.restype = ctypes.c_int32

    # int32_t GetLaserControlSetting(int32_t PS, int32_t Port, int32_t* iSet, double* dSet, char* sSet)
    dll.GetLaserControlSetting.argtypes = [ctypes.c_int32, ctypes.c_int32, ctypes.POINTER(ctypes.c_int32), ctypes.POINTER(ctypes.c_double), ctypes.POINTER(ctypes.c_char)]
    dll.GetLaserControlSetting.restype = ctypes.c_int32

    # int32_t SetLaserControlSetting(int32_t PS, int32_t Port, int32_t iSet, double dSet, char* sSet)
    dll.SetLaserControlSetting.argtypes = [ctypes.c_int32, ctypes.c_int32, ctypes.c_int32, ctypes.c_double, ctypes.POINTER(ctypes.c_char)]
    dll.SetLaserControlSetting.restype = ctypes.c_int32

    # int32_t ClearPIDHistory(int32_t Port)
    dll.ClearPIDHistory.argtypes = [ctypes.c_int32]
    dll.ClearPIDHistory.restype = ctypes.c_int32


# ***********  Other...-functions  *************************************
    # double ConvertUnit(double Val, int32_t uFrom, int32_t uTo)
    dll.ConvertUnit.argtypes = [ctypes.c_double, ctypes.c_int32, ctypes.c_int32]
    dll.ConvertUnit.restype = ctypes.c_double

    # double ConvertDeltaUnit(double Base, double Delta, int32_t uBase, int32_t uFrom, int32_t uTo)
    dll.ConvertDeltaUnit.argtypes = [ctypes.c_double, ctypes.c_double, ctypes.c_int32, ctypes.c_int32, ctypes.c_int32]
    dll.ConvertDeltaUnit.restype = ctypes.c_double


# ***********  Obsolete...-functions  **********************************
    # bool GetReduced(bool R)
    dll.GetReduced.argtypes = [ctypes.c_bool]
    dll.GetReduced.restype = ctypes.c_bool

    # int32_t SetReduced(bool R)
    dll.SetReduced.argtypes = [ctypes.c_bool]
    dll.SetReduced.restype = ctypes.c_int32

    # uint16_t GetScale(uint16_t S)
    dll.GetScale.argtypes = [ctypes.c_uint16]
    dll.GetScale.restype = ctypes.c_uint16

    # int32_t SetScale(uint16_t S)
    dll.SetScale.argtypes = [ctypes.c_uint16]
    dll.SetScale.restype = ctypes.c_int32


    return dll



import unittest
from unittest.mock import patch
import logging
import struct

import qmi
from qmi.instruments.parallax import Parallax_UsbPropeller


class ParallaxUsbPropellerOpenCloseTestCase(unittest.TestCase):

    def setUp(self) -> None:
        qmi.start("parallax_unit_test")
        transport = "serial:COM1"
        self.parallax = qmi.make_instrument("Parallax", Parallax_UsbPropeller, transport)

    def tearDown(self) -> None:
        qmi.stop()

    def test_open_close(self):
        """Test opening and closing the instrument"""
        with patch("serial.Serial") as ser:
            self.parallax.open()
            self.assertTrue(self.parallax.is_open())
            self.parallax.close()
            self.assertFalse(self.parallax.is_open())
            ser.assert_called_once_with(
                "COM1",  # The rest are defaults
                baudrate=2400,
                bytesize=8,
                parity='N',
                rtscts=False,
                stopbits=1.0,
                timeout=0.04
                )


class ParallaxUsbPropellerCommandsTestCase(unittest.TestCase):

    def setUp(self) -> None:
        # suppress logging
        logging.getLogger("qmi.instruments.parallax.usb_propeller").setLevel(logging.CRITICAL)
        self._lead_in = b"!", b"S", b"C"
        self._endline = b"\r"
        qmi.start("parallax_unit_test")
        transport = "serial:/dev/ttyS1"
        self.patcher_open = patch("qmi.core.transport.QMI_Transport.open")
        self.patcher_discard = patch("qmi.core.transport.QMI_SerialTransport.discard_read")
        self.patcher_open.start()
        self.patcher_discard.start()
        self.parallax = qmi.make_instrument("Parallax", Parallax_UsbPropeller, transport)
        self.parallax.open()

    def tearDown(self) -> None:
        self.patcher_open.stop()
        self.patcher_discard.stop()
        patcher_close = patch("qmi.core.transport.QMI_Transport.close")
        patcher_close.start()
        self.parallax.close()
        qmi.stop()
        patcher_close.stop()
        # restore logging
        logging.getLogger("qmi.instruments.parallax.usb_propeller").setLevel(logging.NOTSET)

    def test_get_idn(self):
        """Test getting the QMI instrument ID."""
        expected_vendor = "Parallax"
        expected_model = "USB Propeller"
        expected_version = "1.0"
        expected_command = struct.pack(
                "cccccccc", b"!", b"S", b"C", b"V", b"E", b"R", b"?", self._endline
            )
        with patch("qmi.core.transport.QMI_SerialTransport.write") as write:
            with patch("qmi.core.transport.QMI_SerialTransport.read") as read:
                read.return_value = struct.pack("BBB", 49, 46, 48)  # "1", ".", "0"
                idn = self.parallax.get_idn()

            write.assert_called_with(expected_command)

        self.assertEqual(expected_vendor, idn.vendor)
        self.assertEqual(expected_model, idn.model)
        self.assertIsNone(idn.serial)
        self.assertEqual(expected_version, idn.version)

    def test_set_target_value_with_speed(self):
        """Test setting the target value."""
        target, ramp_speed, channel = 500, 10, 1
        lsb = target & 0xFF  # 7 bits for least significant byte
        msb = target >> 8 & 0xFF  # shift 8 and take next 8 bits for msb
        expected_command = struct.pack(
                "cccBBBBc", b"!", b"S", b"C", channel, ramp_speed, lsb, msb, self._endline
            )

        with patch("qmi.core.transport.QMI_SerialTransport.write") as write:
            self.parallax.set_speed(channel, ramp_speed)
            self.parallax.set_target_value(channel, target)
            write.assert_any_call(expected_command)

    def test_set_target_value_enforces_limits(self):
        """Test setting the target value."""
        expected_pos = [Parallax_UsbPropeller.DEFAULT_MIN, Parallax_UsbPropeller.DEFAULT_MAX]
        # 1. Test minimum value
        target_min, ramp_speed, channel = 300, 10, 1
        lsb = expected_pos[0] & 0xFF  # 7 bits for least significant byte
        msb = expected_pos[0] >> 8 & 0xFF  # shift 8 and take next 8 bits for msb
        expected_command = struct.pack(
                "cccBBBBc", b"!", b"S", b"C", channel, ramp_speed, lsb, msb, self._endline
            )

        with patch("qmi.core.transport.QMI_SerialTransport.write") as write:
            self.parallax.set_speed(channel, ramp_speed)
            self.parallax.set_target_value(channel, target_min)
            write.assert_any_call(expected_command)

        # 2. Test maximum value
        target_max, ramp_speed, channel = 1000, 10, 1
        lsb = expected_pos[1] & 0xFF  # 7 bits for least significant byte
        msb = expected_pos[1] >> 8 & 0xFF  # shift 8 and take next 8 bits for msb
        expected_command = struct.pack(
                "cccBBBBc", b"!", b"S", b"C", channel, ramp_speed, lsb, msb, self._endline
            )

        with patch("qmi.core.transport.QMI_SerialTransport.write") as write:
            self.parallax.set_speed(channel, ramp_speed)
            self.parallax.set_target_value(channel, target_max)
            write.assert_any_call(expected_command)

    def test_set_default_move(self):
        """Test setting the default move."""
        target, channel = 555, 10
        lsb = target & 0xFF  # 7 bits for least significant byte
        msb = target >> 8 & 0xFF  # shift 8 and take next 8 bits for msb
        expected_command = struct.pack(
                "ccccBBBc", b"!", b"S", b"C", b"D", channel, lsb, msb, self._endline
            )

        with patch("qmi.core.transport.QMI_SerialTransport.write") as write:
            self.parallax.set_default_move(channel, target)
            write.assert_any_call(expected_command)

    def test_set_default_move_invalid_values(self):
        """Test setting the default move with invalid values."""
        targets, channel = [0, Parallax_UsbPropeller.MAX_PULSE_WIDTH + 1], 10
        for target in targets:
            with self.assertRaises(ValueError):
                self.parallax.set_default_move(channel, target)

    def test_get_target_values(self):
        """Test get target values return expected values."""
        target = 500
        expected_initial = [0] * Parallax_UsbPropeller.AVAILABLE_CHANNELS
        expected_after = [target + n for n in range(Parallax_UsbPropeller.AVAILABLE_CHANNELS)]

        initial = self.parallax.get_target_values()
        # Assert
        self.assertListEqual(expected_initial, initial)

        # Then set all target values
        with patch("qmi.core.transport.QMI_SerialTransport.write"):
            for channel in range(Parallax_UsbPropeller.AVAILABLE_CHANNELS):
                self.parallax.set_target_value(channel, target + channel)

        after = self.parallax.get_target_values()
        # Assert
        self.assertListEqual(expected_after, after)

    def test_get_value(self):
        """Test get target value returns expected value."""
        expected, channel = 500, 0
        expected_command = struct.pack(
                "ccccccBc", b"!", b"S", b"C", b"R", b"S", b"P", channel, self._endline
            )
        low_bit, high_bit = expected - (expected // 256 * 256), expected // 256
        # Then set all target values
        with patch("qmi.core.transport.QMI_SerialTransport.write") as write:
            with patch("qmi.core.transport.QMI_SerialTransport.read") as read:
                read.return_value = struct.pack("BBB", channel, high_bit, low_bit)
                result = self.parallax.get_value(channel)

            write.assert_called_with(expected_command)

        self.assertEqual(result, expected)

    def test_set_baud_rate(self):
        """Test setting a new baud rate speed."""
        speed = 38400
        expected_command = struct.pack(
                "ccccccBc", b"!", b"S", b"C", b"S", b"B", b"R", 1, self._endline
            )

        with patch("qmi.core.transport.QMI_SerialTransport.write") as write:
            with patch("qmi.core.transport.QMI_SerialTransport.read") as read:
                read.return_value = struct.pack("ccB", b"B", b"R", 1)
                self.parallax.set_baud_rate(speed)

            write.assert_called_with(expected_command)

    def test_set_baud_rate_raises_exception_invalid_speed(self):
        """Test setting an invalid baud rate speed raises an exception."""
        invalid_speeds = (4800, 9600, 19200, 2)
        for speed in invalid_speeds:
            with self.assertRaises(ValueError):
                self.parallax.set_baud_rate(speed)

    def test_set_baud_rate_not_successful_raises_exception(self):
        """Test setting a baud rate speed that was not successful raises an exception."""
        speeds = (1, 0)
        for speed in speeds:
            with patch("qmi.core.transport.QMI_SerialTransport.write"):
                with patch("qmi.core.transport.QMI_SerialTransport.read") as read:
                    read.return_value = struct.pack("ccB", b"B", b"R", int(not speed))
                    with self.assertRaises(AssertionError):
                        self.parallax.set_baud_rate(speed)

    def test_set_software_port_range(self):
        """Test setting a new software port range."""
        port_range = 1
        expected_command = struct.pack(
                "ccccccBc", b"!", b"S", b"C", b"P", b"S", b"S", port_range, self._endline
            )

        with patch("qmi.core.transport.QMI_SerialTransport.write") as write:
            with patch("qmi.core.transport.QMI_SerialTransport.read") as read:
                read.return_value = struct.pack("ccB", b"P", b"M", port_range)
                self.parallax.set_software_port_range(port_range)

            write.assert_called_with(expected_command)

    def test_set_software_port_range_raises_exception_invalid_range(self):
        """Test setting an invalid port range raises an exception."""
        invalid_range = 2
        with self.assertRaises(ValueError):
            self.parallax.set_software_port_range(invalid_range)

    def test_set_software_port_range_not_successful_raises_exception(self):
        """Test setting a software port range that was not successful raises an exception."""
        port_ranges = (1, 0)
        for port_range in port_ranges:
            with patch("qmi.core.transport.QMI_SerialTransport.write"):
                with patch("qmi.core.transport.QMI_SerialTransport.read") as read:
                    read.return_value = struct.pack("ccB", b"P", b"M", int(not port_range))
                    with self.assertRaises(AssertionError):
                        self.parallax.set_software_port_range(port_range)

    def test_disable_servo_channel(self):
        """Test disable servo channel function."""
        channel = 12
        expected_command = struct.pack(
                "ccccccBc", b"!", b"S", b"C", b"P", b"S", b"D", channel, self._endline
            )
        with patch("qmi.core.transport.QMI_SerialTransport.write") as write:
            self.parallax.disable_servo_channel(channel)
            write.assert_called_once_with(expected_command)

    def test_enable_servo_channel(self):
        """Test enable servo channel function."""
        channel = 11
        expected_command = struct.pack(
                "ccccccBc", b"!", b"S", b"C", b"P", b"S", b"E", channel, self._endline
            )
        with patch("qmi.core.transport.QMI_SerialTransport.write") as write:
            self.parallax.enable_servo_channel(channel)
            write.assert_called_once_with(expected_command)

    def test_set_startup_servo_mode(self):
        """Test setting a new startup servo mode."""
        servo_mode = 1
        expected_command = struct.pack(
                "ccccccBc", b"!", b"S", b"C", b"E", b"D", b"D", servo_mode, self._endline
            )

        with patch("qmi.core.transport.QMI_SerialTransport.write") as write:
            with patch("qmi.core.transport.QMI_SerialTransport.read") as read:
                read.return_value = struct.pack("ccB", b"D", b"L", servo_mode)
                self.parallax.set_startup_servo_mode(servo_mode)

            write.assert_called_with(expected_command)

    def test_set_startup_servo_mode_raises_exception_invalid_range(self):
        """Test setting an invalid servo mode raises an exception."""
        invalid_mode = 2
        with self.assertRaises(ValueError):
            self.parallax.set_startup_servo_mode(invalid_mode)

    def test_set_startup_servo_mode_not_successful_raises_exception(self):
        """Test setting a startup servo mode that was not successful raises an exception."""
        servo_modes = (1, 0)
        for servo_mode in servo_modes:
            with patch("qmi.core.transport.QMI_SerialTransport.write") as write:
                with patch("qmi.core.transport.QMI_SerialTransport.read") as read:
                    read.return_value = struct.pack("ccB", b"D", b"L", int(not servo_mode))
                    with self.assertRaises(AssertionError):
                        self.parallax.set_startup_servo_mode(servo_mode)

    def test_clear(self):
        """Test clear a channel function."""
        expected_command = struct.pack(
                "cccccccc", b"!", b"S", b"C", b"L", b"E", b"A", b"R", self._endline
            )
        with patch("qmi.core.transport.QMI_SerialTransport.write") as write:
            with patch("qmi.core.transport.QMI_SerialTransport.read") as read:
                read.return_value = struct.pack("ccc", b"C", b"L", b"R")
                self.parallax.clear()

            write.assert_called_once_with(expected_command)

    def test_clear_assertion_error(self):
        """Test clear a channel function with erroneous response."""
        with patch("qmi.core.transport.QMI_SerialTransport.write"):
            with patch("qmi.core.transport.QMI_SerialTransport.read") as read:
                read.return_value = struct.pack("ccc", b"1", b".", b"0")
                with self.assertRaises(AssertionError):
                    self.parallax.clear()

    def test_move_up_moves_to_max(self):
        """Test that the move_up sets the target value to maximum."""
        channel = 2
        expected_targets = [0] * Parallax_UsbPropeller.AVAILABLE_CHANNELS
        expected_targets[channel] = Parallax_UsbPropeller.DEFAULT_MAX
        with patch("qmi.core.transport.QMI_SerialTransport.write"):
            self.parallax.move_up(channel)
            # Get the target values
            targets = self.parallax.get_target_values()

        self.assertListEqual(expected_targets, targets)

    def test_move_down_moves_to_min(self):
        """Test that the move_down sets the target value to minimum."""
        channel = 3
        expected_targets = [0] * Parallax_UsbPropeller.AVAILABLE_CHANNELS
        expected_targets[channel] = Parallax_UsbPropeller.DEFAULT_MIN
        with patch("qmi.core.transport.QMI_SerialTransport.write"):
            self.parallax.move_down(channel)
            # Get the target values
            targets = self.parallax.get_target_values()

        self.assertListEqual(expected_targets, targets)

    def test_is_moving(self):
        """Test that is_moving returns the correct value."""
        # 1. Not moving, i.o.w. "servo reached target"
        expected, channel = 500, 0
        expected_command = struct.pack(
                "ccccccBc", b"!", b"S", b"C", b"R", b"S", b"P", channel, self._endline
            )
        high_bit, low_bit = expected - (expected // 256 * 256), expected // 256
        # Then set all target values
        with patch("qmi.core.transport.QMI_SerialTransport.write") as write:
            self.parallax.set_target_value(channel, expected)
            with patch("qmi.core.transport.QMI_SerialTransport.read") as read:
                read.return_value = struct.pack("BBB", channel, low_bit, high_bit)
                moving = self.parallax.is_moving(channel)

            write.assert_called_with(expected_command)

        self.assertFalse(moving)

        # 2. Moving
        channel = 1
        expected_command = struct.pack(
                "ccccccBc", b"!", b"S", b"C", b"R", b"S", b"P", channel, self._endline
            )
        not_expected = expected + 1
        high_bit, low_bit = not_expected - (not_expected // 256 * 256), not_expected // 256
        with patch("qmi.core.transport.QMI_SerialTransport.write") as write:
            self.parallax.set_target_value(channel, expected)
            with patch("qmi.core.transport.QMI_SerialTransport.read") as read:
                read.return_value = struct.pack("BBB", channel, low_bit, high_bit)
                moving = self.parallax.is_moving(channel)

            write.assert_called_with(expected_command)

        self.assertTrue(moving)

    def test_set_max(self):
        """Test that setting new maximum value on a channels works."""
        new_max = 1000
        channel = 2
        expected_targets = [0] * Parallax_UsbPropeller.AVAILABLE_CHANNELS
        expected_targets[channel] = new_max

        self.parallax.set_max(channel, new_max)
        with patch("qmi.core.transport.QMI_SerialTransport.write"):
            self.parallax.move_up(channel)
            # Get the target values
            targets = self.parallax.get_target_values()

        self.assertListEqual(expected_targets, targets)

    def test_set_max_excepts(self):
        """Test that setting new maximum value outside valid range excepts."""
        invalid_maxes = [100, 1050]
        channel = 4

        for new_max in invalid_maxes:
            with self.assertRaises(ValueError):
                self.parallax.set_max(channel, new_max)

    def test_set_min(self):
        """Test that setting new minimum value on a channels works."""
        new_min = 100
        channel = 3
        expected_targets = [0] * Parallax_UsbPropeller.AVAILABLE_CHANNELS
        expected_targets[channel] = new_min

        self.parallax.set_min(channel, new_min)
        with patch("qmi.core.transport.QMI_SerialTransport.write"):
            self.parallax.move_down(channel)
            # Get the target values
            targets = self.parallax.get_target_values()

        self.assertListEqual(expected_targets, targets)

    def test_set_min_excepts(self):
        """Test that setting new minimum value outside valid range excepts."""
        invalid_mins = [-1, 1000]
        channel = 5

        for new_min in invalid_mins:
            with self.assertRaises(ValueError):
                self.parallax.set_min(channel, new_min)

    def test_set_wrong_channel_excepts(self):
        """Test a set command with wrong channel numbers and see that it excepts"""
        wrong_channels = (-1, Parallax_UsbPropeller.AVAILABLE_CHANNELS)
        for ch in wrong_channels:
            with self.assertRaises(ValueError):
                self.parallax.set_target_value(ch, 0)


if __name__ == '__main__':
    unittest.main()



#! /usr/bin/env python3

"""
Test shutdown of QMI with instrument still open.

This program does not use the unittest framework because it tests
for problems that are difficult to trigger in the context of
a unit test.

A specific scenario is tested:
  - Application starts QMI.
  - Application creates a QMI instrument instance and opens the instrument.
  - Application leaks a reference to the instrument proxy into a global variable.
  - Application stops QMI (with instrument still open).
  - Application exits.

This is tricky is several ways:
  - The QMI context is stopped while an RPC object is still active.
  - The QMI instrument instance is eventually cleaned up while instrument still open.
  - The global reference to the instrument proxy may hold up cleanup of the
    QMI context until after the logging system has already shut down.
  - During cleanup, QMI may try to log while the logging system is already down
    and thus trigger mayhem.
"""

import socket
import threading
import time

import qmi
from qmi.instruments.cobolt.laser_06_01 import Cobolt_Laser_06_01


_global_proxies = []


def _start_server():
    """Start a background TCP server for the instrument to connect to."""

    global _server_thread
    global _server_port

    # Create server socket.
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

    # Bind to arbitrary port.
    sock.bind(("127.0.0.1", 0))
    (host, _server_port) = sock.getsockname()

    sock.listen(1)
    print("Dummy server waiting for TCP connection on port", _server_port)

    # Start background thread.
    _server_thread = threading.Thread(target=_run_server, args=(sock,))
    _server_thread.start()


def _run_server(sock):

    global _server_conn

    # Within background thread, wait for incoming connection.
    sock.settimeout(1.0)
    while True:
        try:
            (conn, addr) = sock.accept()
            break
        except (TimeoutError, socket.timeout):
            if not threading.main_thread().is_alive():
                print("Dummy server aborting")
                sock.close()
                return
        
    print("Dummy server got incoming TCP connection")
    sock.close()

    # Keep reference to connection socket to keep it open.
    _server_conn = conn

    # End background thread.


def main():

    # Start background TCP server.
    _start_server()

    # Start QMI.
    qmi.start("test_context")

    # Create laser instance.
    transport = "tcp:127.0.0.1:{}".format(_server_port)
    laser = qmi.make_instrument("laser", Cobolt_Laser_06_01, transport)

    # Open instrument.
    laser.open()

    # Leak the instrument proxy to a global variable.
    _global_proxies.append(laser)

    # Clean up background thread.
    _server_thread.join()

    # Wait for dramatic effect.
    time.sleep(1)

    # Stop QMI.
    qmi.stop()


if __name__ == "__main__":
    main()




"""Test for the Bristol 871A driver."""
import struct
import logging

from math import isnan

from unittest import TestCase
from unittest.mock import MagicMock
from unittest.mock import patch
from unittest.mock import call

from typing import cast

from dataclasses import dataclass


from qmi.core.transport import QMI_Transport
from qmi.core.scpi_protocol import ScpiProtocol
from qmi.instruments.bristol.bristol_871a import Bristol_871A
from qmi.instruments.bristol.bristol_871a import _ReaderThread
from qmi.instruments.bristol.bristol_871a import Measurement
from qmi.core.exceptions import QMI_InstrumentException


# Disable all logging
logging.disable(logging.CRITICAL)


@dataclass
class TestMetaReaderThread:
    """Test meta data."""

    transport: MagicMock
    super: MagicMock
    queue: MagicMock()
    reader: _ReaderThread


@dataclass
class TestMeta871a:
    """Test meta data."""

    serial: MagicMock
    serial_str: MagicMock
    scpi_transport: MagicMock
    scpi_protocol: MagicMock
    scpi_str: MagicMock
    name: MagicMock
    super: MagicMock
    reader: MagicMock
    queue: MagicMock()
    instr: Bristol_871A


class TestReaderThread(TestCase):
    def setUp(self):
        mock_transport = MagicMock(spec=QMI_Transport)
        mock_super = MagicMock()
        mock_queue = MagicMock()

        reader = _ReaderThread(mock_transport, mock_queue)

        self._patcher_super = patch(
            "qmi.instruments.bristol.bristol_871a.super", mock_super
        )
        self._patcher_super.start()

        self.meta = TestMetaReaderThread(
            transport=mock_transport,
            super=mock_super,
            queue=mock_queue,
            reader=reader,
        )

    def tearDown(self):
        self.meta = None
        self._patcher_super.stop()

    def test_init(self):
        """_ReaderThread.__init__(), happy flow"""
        self.assertEqual(self.meta.reader._transport, self.meta.transport)
        self.assertEqual(self.meta.reader._queue, self.meta.queue)

    class TestException(Exception):
        pass

    def test_run(self):
        """_ReaderThread.run(), happy flow."""
        # using TestException to exit the while loop.
        mock_measurement = MagicMock()

        self.meta.queue.append = MagicMock(side_effect=self.TestException)
        self.meta.reader._read_measurement = MagicMock(side_effect=[None, mock_measurement])
        with self.assertRaises(self.TestException):
            self.meta.reader.run()

        self.meta.reader._read_measurement.assert_called_with()
        self.meta.queue.append.assert_called_once_with(mock_measurement)

    def test_read_measurement(self):
        """_ReaderThread._read_measurement(), happy flow."""
        mock_wavelength = 0.0
        mock_power = 1.001
        mock_status = 5
        mock_index = 6

        mock_data_0 = b"\x00"
        mock_data_1 = b"\x7e"
        mock_data_2 = struct.pack(
            "<dfII",
            mock_wavelength,
            mock_power,
            mock_status,
            mock_index,
        )
        mock_data_3 = b"0x7d"
        mock_time = MagicMock()
        self.meta.transport.read_until_timeout = MagicMock(
            side_effect=[mock_data_0, mock_data_1, mock_data_2, mock_data_3]
        )

        with patch("qmi.instruments.bristol.bristol_871a.time.time", mock_time):
            rt_val = self.meta.reader._read_measurement()

        self.assertEqual(rt_val.timestamp, mock_time())
        self.assertTrue(isnan(rt_val.wavelength))
        self.assertAlmostEqual(rt_val.power, mock_power, 3)
        self.assertEqual(rt_val.status, mock_status)
        self.assertEqual(rt_val.index, mock_index)

    def test_read_measurement_corrupt(self):
        """_ReaderThread._read_measurement(), data corrupt."""
        # corruption occurs due to a second start token within a message.
        # the driver should clear the buffer and parse the new message.
        mock_wavelength = 0.0
        mock_power = 1.001
        mock_status = 5
        mock_index = 6

        mock_data_1 = b"\x7e"
        mock_data_2 = b"\x7e"
        mock_data_3 = struct.pack(
            "<dfII",
            mock_wavelength,
            mock_power,
            mock_status,
            mock_index,
        )
        mock_data_4 = b"0x7d"

        mock_time = MagicMock()
        self.meta.transport.read_until_timeout = MagicMock(
            side_effect=[mock_data_1, mock_data_2, mock_data_3, mock_data_4]
        )

        with patch("qmi.instruments.bristol.bristol_871a.time.time", mock_time):
            rt_val = self.meta.reader._read_measurement()

        self.assertEqual(rt_val.timestamp, mock_time())
        self.assertTrue(isnan(rt_val.wavelength))
        self.assertAlmostEqual(rt_val.power, mock_power, 3)
        self.assertEqual(rt_val.status, mock_status)
        self.assertEqual(rt_val.index, mock_index)

    def test_read_measurement_shutdown_requested(self):
        """See that the function returns early if shutdown requested."""
        with patch("qmi.instruments.bristol.bristol_871a.time.time", MagicMock()):
            self.meta.reader._shutdown_requested = True
            rt_val = self.meta.reader._read_measurement()

        self.assertIsNone(rt_val)


class TestBristol_871A(TestCase):
    """Testcase for the Bristol_871A class."""

    def setUp(self):
        mock_transport_serial = MagicMock(spec=QMI_Transport)
        mock_transport_serial_str = MagicMock()
        mock_transport_scpi = MagicMock(spec=QMI_Transport)
        mock_transport_scpi_str = MagicMock()
        mock_name = MagicMock()
        mock_scpi = MagicMock(spec=ScpiProtocol)
        mock_super = MagicMock()
        mock_reader = MagicMock(spec=_ReaderThread)
        mock_queue = MagicMock()

        def se_create_transport(transport):
            if transport == mock_transport_scpi_str:
                return mock_transport_scpi
            elif transport == mock_transport_serial_str:
                return mock_transport_serial

        with patch(
            "qmi.instruments.bristol.bristol_871a.create_transport",
            side_effect=se_create_transport,
        ), patch("qmi.instruments.bristol.bristol_871a.ScpiProtocol", mock_scpi), patch(
            "qmi.instruments.bristol.bristol_871a._ReaderThread", mock_reader
        ), patch(
            "qmi.instruments.bristol.bristol_871a.collections.deque", mock_queue
        ):
            instr = Bristol_871A(
                MagicMock(),
                mock_name,
                mock_transport_scpi_str,
                mock_transport_serial_str,
            )

        self._patcher_super = patch(
            "qmi.instruments.bristol.bristol_871a.super", mock_super
        )
        self._patcher_super.start()

        self.meta = TestMeta871a(
            serial=mock_transport_serial,
            scpi_protocol=mock_scpi,
            scpi_transport=mock_transport_scpi,
            serial_str=mock_transport_serial_str,
            scpi_str=mock_transport_scpi_str,
            name=mock_name,
            super=mock_super,
            reader=mock_reader,
            queue=mock_queue,
            instr=cast(Bristol_871A, instr),
        )

    def tearDown(self):
        self.meta = None
        self._patcher_super.stop()

    def test_init(self):
        """Bristol_871A.__init__(), happy flow"""
        self.meta.queue.assert_called_once_with(
            maxlen=self.meta.instr.DEFAULT_QUEUE_SIZE
        )
        self.assertEqual(self.meta.instr._serial_transport, self.meta.serial)
        self.meta.scpi_protocol.assert_called_once_with(self.meta.scpi_transport)
        self.assertEqual(self.meta.instr._scpi_protocol, self.meta.scpi_protocol())
        self.meta.reader.assert_called_once_with(self.meta.serial, self.meta.queue())
        self.meta.reader().start.assert_called_once_with()

    def test_init_value_error(self):
        """Bristol_871A.__init__(), value error handling"""
        with self.assertRaises(ValueError):
            Bristol_871A(
                object(),
                "fouteboel",
                None,
                None,
            )

    def test_open(self):
        """Bristol_871A.open(), happy flow."""
        self.meta.instr._scpi_handshake = MagicMock()

        self.meta.instr.open()
        self.meta.scpi_transport.open.assert_called_once_with()
        self.meta.serial.open.assert_called_once_with()
        self.meta.super().open.assert_called_once_with()
        self.meta.instr._scpi_handshake.assert_called_once_with()

    def test_open_scpi_handshake_exception(self):
        """Bristol_871A.open(), scpi handshake exception handling."""
        self.meta.instr._scpi_handshake = MagicMock(side_effect=Exception)
        with self.assertRaises(Exception):
            self.meta.instr.open()
        self.meta.scpi_transport.close.assert_called_once_with()

    def test_close(self):
        """Bristol_871A.close(), happy flow."""
        self.meta.instr._check_is_open = MagicMock()

        self.meta.instr.close()
        self.meta.instr._check_is_open.assert_called_once_with()
        self.meta.reader().shutdown.assert_called_once_with()
        self.meta.reader().join.assert_called_once_with()
        self.meta.super().close.assert_called_once_with()
        self.meta.serial.close.assert_called_once_with()
        self.meta.scpi_transport.close.assert_called_once_with()

    def test_write_scpi(self):
        """Bristol_871A._write_scpi(), happy flow."""
        mock_cmd = MagicMock()
        self.meta.instr._write_scpi(mock_cmd)
        self.meta.scpi_protocol().write.assert_called_once_with(mock_cmd)

    def test_ask_scpi(self):
        """Bristol_871A._ask_scpi(), happy flow."""
        mock_cmd = MagicMock()
        rt_val = self.meta.instr._ask_scpi(mock_cmd)
        self.meta.scpi_protocol().ask.assert_called_once_with(mock_cmd)
        self.assertEqual(rt_val, self.meta.scpi_protocol().ask().rstrip("\r\n"))

    def test_reset(self):
        """Bristol_871A.reset(), happy flow."""
        self.meta.instr._write_scpi = MagicMock()
        self.meta.instr.reset()
        self.meta.instr._write_scpi.assert_called_once_with("*RST")

    def test_get_idn(self):
        """Bristol_871A.get_idn(), happy flow."""
        mock_resp = "VENDOR,MODEL,SERIAL,VERSION"

        self.meta.instr._ask_scpi = MagicMock(return_value=mock_resp)

        rt_val = self.meta.instr.get_idn()
        self.assertEqual(rt_val.vendor, "VENDOR")
        self.assertEqual(rt_val.model, "MODEL")
        self.assertEqual(rt_val.serial, "SERIAL")
        self.assertEqual(rt_val.version, "VERSION")

    def test_get_idn_invalid_response(self):
        """Bristol_871A.get_idn(), invalid response handling."""
        mock_resp = ""
        self.meta.instr._ask_scpi = MagicMock(return_value=mock_resp)
        with self.assertRaises(QMI_InstrumentException):
            self.meta.instr.get_idn()

    def test_scpi_handshake(self):
        """Bristol_871A.scpi_handshake(), happy flow."""
        mock_resp_1 = b"Bristol Instruments"
        mock_resp_2 = b"BRISTOL"

        self.meta.scpi_transport.read_until = MagicMock(
            side_effect=[mock_resp_1, mock_resp_2]
        )
        self.meta.instr._write_scpi = MagicMock()

        self.meta.instr._scpi_handshake()

        self.meta.scpi_transport.read_until.assert_has_calls(
            [
                call(
                    message_terminator=b"\n", timeout=self.meta.instr.RESPONSE_TIMEOUT
                ),
                call(
                    message_terminator=b"\n", timeout=self.meta.instr.RESPONSE_TIMEOUT
                ),
            ]
        )
        self.meta.instr._write_scpi.assert_called_once_with("*IDN?")

    def test_scpi_handshake_no_connection(self):
        """Bristol_871A.scpi_handshake(), no connection handling."""
        mock_resp = b"no connections available"

        self.meta.scpi_transport.read_until = MagicMock(return_value=mock_resp)

        with self.assertRaises(QMI_InstrumentException):
            self.meta.instr._scpi_handshake()

        self.meta.scpi_transport.read_until.assert_has_calls(
            [
                call(
                    message_terminator=b"\n", timeout=self.meta.instr.RESPONSE_TIMEOUT
                ),
            ]
        )

    def test_scpi_handshake_invalid_response(self):
        """Bristol_871A.scpi_handshake(), invalid response handling."""
        mock_resp_1 = b"Bristol Instruments"
        mock_resp_2 = b"*IDN?"

        self.meta.scpi_transport.read_until = MagicMock(
            side_effect=[mock_resp_1, mock_resp_2]
        )
        self.meta.instr._write_scpi = MagicMock()

        with self.assertRaises(QMI_InstrumentException):
            self.meta.instr._scpi_handshake()

        self.meta.scpi_transport.read_until.assert_has_calls(
            [
                call(
                    message_terminator=b"\n", timeout=self.meta.instr.RESPONSE_TIMEOUT
                ),
                call(
                    message_terminator=b"\n", timeout=self.meta.instr.RESPONSE_TIMEOUT
                ),
            ]
        )
        self.meta.instr._write_scpi.assert_called_once_with("*IDN?")

    def test_parse_int(self):
        """Bristol_871A._parse_int(), happy flow."""
        mock_resp = MagicMock()
        mock_cmd = MagicMock()
        rt_val = self.meta.instr._parse_int(mock_resp, mock_cmd)
        self.assertEqual(rt_val, int(mock_resp))

    def test_parse_int_invalid_response(self):
        """Bristol_871A._parse_int(), invalid response handling."""
        mock_resp = "INVALID"
        mock_cmd = MagicMock()
        with self.assertRaises(QMI_InstrumentException):
            self.meta.instr._parse_int(mock_resp, mock_cmd)

    def _test_is_valid_measurement(self, status, wavelength, is_valid):
        mock_value = Measurement(
            status=status, wavelength=wavelength, timestamp=0, index=0, power=0
        )
        rt_val = self.meta.instr.is_valid_measurement(mock_value)
        self.assertEqual(rt_val, is_valid)

    def test_is_valid_measurement(self):
        """Bristol_871A.is_valid_measurement(), happy flow."""
        self._test_is_valid_measurement(0x4, 1, True)
        self._test_is_valid_measurement(0x4, 0, False)
        self._test_is_valid_measurement(0x0, 1, False)

    def test_read_measurement(self):
        """Bristol_871A.read_measurement(), happy flow."""
        mock_wavelength = 0.0
        mock_power = 1.001
        mock_status = 5
        mock_index = 6
        mock_resp = f"{mock_index},{mock_status},{mock_wavelength},{mock_power}"
        mock_time = MagicMock()
        self.meta.instr._ask_scpi = MagicMock(return_value=mock_resp)

        with patch("qmi.instruments.bristol.bristol_871a.time.time", mock_time):
            rt_val = self.meta.instr.read_measurement()

        self.assertEqual(rt_val.timestamp, mock_time())
        self.assertEqual(rt_val.index, mock_index)
        self.assertEqual(rt_val.status, mock_status)
        self.assertTrue(isnan(rt_val.wavelength))
        self.assertEqual(rt_val.power, mock_power)

        self.meta.instr._ask_scpi.assert_called_once_with(":READ:ALL?")

    def test_read_measurement_invalid_response(self):
        """Bristol_871A.read_measurement(), invalid response handling."""
        mock_resp = ""
        mock_time = MagicMock()
        self.meta.instr._ask_scpi = MagicMock(return_value=mock_resp)

        with patch("qmi.instruments.bristol.bristol_871a.time.time", mock_time):
            with self.assertRaises(QMI_InstrumentException):
                self.meta.instr.read_measurement()

        self.meta.instr._ask_scpi.assert_called_once_with(":READ:ALL?")

    def test_read_measurement_value_error(self):
        """Bristol_871A.read_measurement(), value error handling."""
        mock_resp = "INVALID,INVALID,INVALID,INVALID"
        mock_time = MagicMock()
        self.meta.instr._ask_scpi = MagicMock(return_value=mock_resp)

        with patch("qmi.instruments.bristol.bristol_871a.time.time", mock_time):
            with self.assertRaises(QMI_InstrumentException):
                self.meta.instr.read_measurement()

        self.meta.instr._ask_scpi.assert_called_once_with(":READ:ALL?")

    def test_calibrate(self):
        """Bristol_871A.calibrate(), happy flow."""
        self.meta.instr._write_scpi = MagicMock()
        with patch("qmi.instruments.bristol.bristol_871a.time.sleep", MagicMock()):
            self.meta.instr.calibrate()
        self.meta.instr._write_scpi.assert_called_once_with(":SENS:CALI")

    def _test_get_generic(self, function_name, cmd):
        mock_resp = MagicMock()
        self.meta.instr._ask_scpi = MagicMock(return_value=mock_resp)
        rt_val = getattr(self.meta.instr, function_name)()
        self.assertEqual(rt_val, mock_resp)
        self.meta.instr._ask_scpi.assert_called_once_with(cmd)

    def test_get_auto_calibration_method(self):
        """Bristol_871A.get_auto_calibration_method(), happy flow."""
        self._test_get_generic("get_auto_calibration_method", ":SENS:CALI:METH?")

    def test_get_trigger_method(self):
        """Bristol_871A.get_trigger_method(), happy flow."""
        self._test_get_generic("get_trigger_method", ":TRIG:SEQ:METH?")

    def _test_get_int(self, function_name, cmd):
        mock_resp = MagicMock()
        mock_int = MagicMock()
        self.meta.instr._ask_scpi = MagicMock(return_value=mock_resp)
        self.meta.instr._parse_int = MagicMock(return_value=mock_int)
        rt_val = getattr(self.meta.instr, function_name)()
        self.assertEqual(rt_val, mock_int)
        self.meta.instr._ask_scpi.assert_called_once_with(cmd)
        self.meta.instr._parse_int.assert_called_once_with(mock_resp, cmd)

    def test_get_auto_calibration_temperature(self):
        """Bristol_871A.get_auto_calibration_temperature(), happy flow."""
        self._test_get_int("get_auto_calibration_temperature", ":SENS:CALI:TEMP?")

    def test_get_auto_calibration_time(self):
        """Bristol_871A.get_auto_calibration_time(), happy flow."""
        self._test_get_int("get_auto_calibration_time", ":SENS:CALI:TIM?")

    def test_get_condition(self):
        """Bristol_871A.get_condition(), happy flow."""
        self._test_get_int("get_condition", ":STAT:QUES:COND?")

    def test_get_trigger_rate(self):
        """Bristol_871A.get_trigger_rate(), happy flow."""
        self._test_get_int("get_trigger_rate", ":TRIG:SEQ:RATE:ADJ?")

    def _test_set_str(self, function_name, cmd, value):
        self.meta.instr._write_scpi = MagicMock()
        getattr(self.meta.instr, function_name)(value)
        self.meta.instr._write_scpi.assert_called_once_with(cmd % value.upper())

    def test_set_auto_calibration_method(self):
        """Bristol_871A.set_auto_calibration_method(), happy flow."""
        self._test_set_str("set_auto_calibration_method", ":SENS:CALI:METH %s", "OFF")
        self._test_set_str("set_auto_calibration_method", ":SENS:CALI:METH %s", "TIME")
        self._test_set_str("set_auto_calibration_method", ":SENS:CALI:METH %s", "TEMP")

    def test_set_trigger_method(self):
        """Bristol_871A.set_trigger_method(), happy flow."""
        self._test_set_str("set_trigger_method", ":TRIG:SEQ:METH %s", "INT")
        self._test_set_str("set_trigger_method", ":TRIG:SEQ:METH %s", "FALL")
        self._test_set_str("set_trigger_method", ":TRIG:SEQ:METH %s", "RISE")

    def test_set_auto_calibration_method_value_error(self):
        """Bristol_871A.set_auto_calibration_method(), value error handling."""
        with self.assertRaises(ValueError):
            self._test_set_str("set_auto_calibration_method", ":SENS:CALI:METH %s", "")

    def test_set_trigger_method_value_error(self):
        """Bristol_871A.set_trigger_method(), value error handling."""
        with self.assertRaises(ValueError):
            self._test_set_str("set_trigger_method", ":TRIG:SEQ:METH %s", "")

    def _test_set_int(self, function_name, cmd, value):
        self.meta.instr._write_scpi = MagicMock()
        getattr(self.meta.instr, function_name)(value)
        try:
            self.meta.instr._write_scpi.assert_called_once_with(cmd % value)
        except TypeError:
            self.meta.instr._write_scpi.assert_called_once_with(cmd)

    def test_set_auto_calibration_temperature(self):
        """Bristol_871A.set_auto_calibration_temperature(), happy flow."""
        self._test_set_int("set_auto_calibration_temperature", ":SENS:CALI:TEMP %d", 1)
        self._test_set_int("set_auto_calibration_temperature", ":SENS:CALI:TEMP %d", 50)

    def test_set_auto_calibration_time(self):
        """Bristol_871A.set_auto_calibration_time(), happy flow."""
        self._test_set_int("set_auto_calibration_time", ":SENS:CALI:TIM %d", 5)
        self._test_set_int("set_auto_calibration_time", ":SENS:CALI:TIM %d", 1440)

    def test_set_trigger_rate(self):
        """Bristol_871A.set_trigger_rate(), happy flow."""
        self._test_set_int("set_trigger_rate", ":TRIG:SEQ:RATE:ADJ", 0)
        self._test_set_int("set_trigger_rate", ":TRIG:SEQ:RATE %d", 20)
        self._test_set_int("set_trigger_rate", ":TRIG:SEQ:RATE %d", 50)
        self._test_set_int("set_trigger_rate", ":TRIG:SEQ:RATE %d", 100)
        self._test_set_int("set_trigger_rate", ":TRIG:SEQ:RATE %d", 250)
        self._test_set_int("set_trigger_rate", ":TRIG:SEQ:RATE %d", 500)
        self._test_set_int("set_trigger_rate", ":TRIG:SEQ:RATE %d", 1000)

    def test_set_auto_calibration_temperature_value_error(self):
        """Bristol_871A.set_auto_calibration_temperature(), value error handling."""
        with self.assertRaises(ValueError):
            self._test_set_int(
                "set_auto_calibration_temperature", ":SENS:CALI:TEMP %d", 0
            )

    def test_set_auto_calibration_time_value_error(self):
        """Bristol_871A.set_auto_calibration_time(), value error handling."""
        with self.assertRaises(ValueError):
            self._test_set_int("set_auto_calibration_time", ":SENS:CALI:TIM %d", 0)

    def test_set_trigger_rate_value_error(self):
        """Bristol_871A.set_trigger_rate(), value error handling."""
        with self.assertRaises(ValueError):
            self._test_set_int("set_trigger_rate", ":TRIG:SEQ:RATE %d", -1)

    def test_memory_start(self):
        """Bristol_871A.memory_start(), happy flow."""
        self.meta.instr._write_scpi = MagicMock()
        self.meta.instr.memory_start()
        self.meta.instr._write_scpi.assert_has_calls(
            [
                call(":MMEM:INIT"),
                call(":MMEM:OPEN"),
            ]
        )

    def test_memory_stop(self):
        """Bristol_871A.memory_stop(), happy flow."""
        self.meta.instr._write_scpi = MagicMock()
        self.meta.instr.memory_stop()
        self.meta.instr._write_scpi.assert_called_once_with(
            ":MMEM:CLOSE",
        )

    def test_get_memory_contents(self):
        """Bristol_871A.get_memory_contents(), happy flow."""
        mock_wavelength = 0.0
        mock_power = 1.001
        mock_status = 5
        mock_index = 6
        mock_resp = struct.pack(
            "<dfII", mock_wavelength, mock_power, mock_status, mock_index
        )
        mock_time = MagicMock()
        self.meta.instr._write_scpi = MagicMock()
        self.meta.scpi_protocol().read_binary_data = MagicMock(return_value=mock_resp)

        with patch("qmi.instruments.bristol.bristol_871a.time.time", mock_time):
            rt_val = self.meta.instr.get_memory_contents()

        self.meta.scpi_protocol().read_binary_data.assert_called_once_with()
        self.assertEqual(rt_val[0].timestamp, mock_time())
        self.assertTrue(isnan(rt_val[0].wavelength))
        self.assertAlmostEqual(rt_val[0].power, mock_power, 3)
        self.assertEqual(rt_val[0].status, mock_status)
        self.assertEqual(rt_val[0].index, mock_index)

    def test_get_memory_contents_invalid_response(self):
        """Bristol_871A.get_memory_contents(), invalid response handling."""
        mock_resp = [
            MagicMock()
        ]  # data length should be modules of 20, data length 1 is not and should result in an error.
        self.meta.instr._write_scpi = MagicMock()
        self.meta.scpi_protocol().read_binary_data = MagicMock(return_value=mock_resp)

        with self.assertRaises(QMI_InstrumentException):
            self.meta.instr.get_memory_contents()

    def test_get_streaming_measurements(self):
        """Bristol_871A.get_streaming_measurements(), happy flow."""
        with patch("qmi.instruments.bristol.bristol_871a.len", return_value=1):
            rt_val = self.meta.instr.get_streaming_measurements()
        self.meta.queue().popleft.assert_called_once_with()
        self.assertEqual(rt_val[0], self.meta.queue().popleft())



"""
Tektronix Inc.; function and waveform generators, frequency counters

The qmi.instruments.tektronix package provides support for:
- AFG31000 series arbitrary function generator
- AWG 5014[C] arbitrary waveform generator
- FCA3000 and compatible frequency counters
"""
from qmi.instruments.tektronix.awg5014 import Tektronix_Awg5014
from qmi.instruments.tektronix.afg31000 import Waveform, BurstMode, TriggerEdge
# Alternative, QMI naming convention approved names
from qmi.instruments.tektronix.afg31000 import Tektronix_AFG31000 as Tektronix_Afg31000
from qmi.instruments.tektronix.fca3000 import Tektronix_FCA3000 as Tektronix_Fca3000



"""Instrument driver for the Quantum Composers 9530 pulse generator."""

import enum
import logging
from typing import Tuple, Union

from qmi.core.context import QMI_Context
from qmi.core.exceptions import QMI_InstrumentException
from qmi.core.instrument import QMI_Instrument, QMI_InstrumentIdentification
from qmi.core.rpc import rpc_method
from qmi.core.transport import create_transport


# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)


class RefClkSource(enum.Enum):
    """Reference clock sources.

    Attributes:
        INTERNAL: Internal clock source.
        EXTERNAL: External clock input.
        EXTPLL:   External clock signal with PLL.
    """
    INTERNAL = "INT"
    EXTERNAL = "EXT"
    EXTPLL = "XPL"


class PulseMode(enum.Enum):
    """Pulse generator modes.

    Attributes:
        NORMAL: Generate continuous pulse sequence (every input event causes an output pulse).
        SINGLE: Generate just one pulse after the output is enabled.
        BURST:  Generate a fixed number of pulses after the output is enabled; then suppress further pulses.
        DUTYCYCLE: Keep alternating between generating a fixed number of pulses and suppressing
                   a fixed number of pulses.
    """
    NORMAL = "NORM"
    SINGLE = "SING"
    BURST = "BURS"
    DUTYCYCLE = "DCYC"


class TriggerMode(enum.Enum):
    """Trigger modes."""
    DISABLED = "DIS"
    ENABLED = "TRIG"
    DUAL = "DUAL"


class TriggerEdge(enum.Enum):
    """Trigger on rising or falling edge."""
    RISING = "RIS"
    FALLING = "FALL"


class OutputDriver(enum.Enum):
    """Output driver type.

    Attributes:
        TTL: Fast driver with fixed 4 Volt output level (into 1 kOhm).
        ADJUSTABLE: Slower driver with programmable output level.
    """
    TTL = "TTL"
    ADJUSTABLE = "ADJ"


# Table for decoding error responses.
_ERROR_MESSAGES = {
    1: "Incorrect prefix",
    2: "Missing command keyword",
    3: "Invalid command keyword",
    4: "Missing parameter",
    5: "Invalid parameter",
    6: "Query only, command needs a question mark",
    7: "Invalid query, command does not have a query form",
    8: "Command unavailable in current system state"
}


class QuantumComposers_PulseGenerator9530(QMI_Instrument):
    """Instrument driver for the Quantum Composers 9530 pulse generator.

    This driver can be used to communicate with the instrument via USB or
    Ethernet (TCP), depending on the specified transport descriptor.
    """

    # Instrument should respond within 2 seconds.
    RESPONSE_TIMEOUT = 2.0

    def __init__(self, context: QMI_Context, name: str, scpi_transport: str) -> None:
        """Initialize driver.

        Depending on the transport descriptor, the instrument may be accessed
        either via USB or via Ethernet. For example::
          "tcp:172.16.3.83:2101"
          or "serial:/dev/serial/by-id/usb-FTDI_FT232R_..."

        Parameters:
            context: QMI context that will manage this RPC object.
            name: Unique name for this RPC object instance.
            scpi_transport: QMI transport descriptor to connect to the instrument.
        """

        # NOTE: The instrument uses an SCPI-like command language, but generates
        # non-compliant response messages. For this reason, the ScpiProtocol
        # class can not be used. Instead the command interaction is implemented
        # within this driver.

        super().__init__(context, name)
        default_attributes = {
            "port": 2101,      # default TCP port number for Digi Connect module
            "baudrate": 38400  # default baud rate when connected via USB
        }
        self._transport = create_transport(scpi_transport, default_attributes=default_attributes)

    @rpc_method
    def open(self) -> None:
        _logger.info("[%s] Opening connection to instrument", self._name)
        self._transport.open()
        super().open()

    @rpc_method
    def close(self) -> None:
        _logger.info("[%s] Closing connection to instrument", self._name)
        super().close()
        self._transport.close()

    @staticmethod
    def _parse_error_response(response: str) -> str:
        """Decode an error response from the instrument and return a human-readable error message."""

        # In case of an error, the device responds with "?<N>"
        # where <N> is a a numeric error code.
        if response.startswith("?"):
            try:
                errcode = int(response[1:])
                return _ERROR_MESSAGES[errcode]
            except ValueError:
                # Invalid error response format; ignore.
                pass
            except KeyError:
                # Unknown error code; ignore.
                pass
        return repr(response)

    @staticmethod
    def _parse_bool(response: str) -> bool:
        """Parse a boolean response from the instrument."""
        response = response.strip()
        if response == "0":
            return False
        if response == "1":
            return True
        raise QMI_InstrumentException(f"Unexpected response {response!r} while expecting boolean")

    @staticmethod
    def _parse_int(response: str) -> int:
        """Parse an integer response from the instrument."""
        try:
            return int(response.strip())
        except ValueError:
            # pylint: disable=raise-missing-from
            raise QMI_InstrumentException(f"Unexpected response {response!r} while expecting integer")

    @staticmethod
    def _parse_float(response: str) -> float:
        """Parse a floating point response from the instrument."""
        try:
            return float(response.strip())
        except ValueError:
            # pylint: disable=raise-missing-from
            raise QMI_InstrumentException(f"Unexpected response {response!r} while expecting float")

    def _ask(self, cmd: str) -> str:
        """Send query to instrument and return response from instrument.

        Parameters:
            cmd: SCPI command string.

        Returns:
            Response message with message terminator removed.
        """

        # Send command.
        raw_cmd = cmd.encode("ascii") + b"\r\n"
        self._transport.write(raw_cmd)

        # Read response.
        raw_response = self._transport.read_until(
            message_terminator=b"\n",
            timeout=self.RESPONSE_TIMEOUT)
        return raw_response.rstrip().decode("ascii")

    def _cmd(self, cmd: str) -> None:
        """Send a (non-query) command to the instrument.

        Parameters:
            cmd: SCPI command string.
        """

        # Send command and read response.
        response = self._ask(cmd)

        # Check that command was accepted.
        response = response.strip()
        if response.lower() != "ok":
            _logger.error("[%s] Error from instrument, cmd=%r, response=%r", self._name, cmd, response)
            errmsg = self._parse_error_response(response)
            raise QMI_InstrumentException(f"Error response from instrument: {errmsg}")

    @staticmethod
    def _check_range(value: Union[int, float],
                     minval: Union[int, float, None],
                     maxval: Union[int, float, None]
                     ) -> None:
        """Check that the parameter value is in the allowed range, otherwise raise ValueError."""
        if (((minval is not None) and (value < minval))
                or ((maxval is not None) and (value > maxval))):
            raise ValueError("Parameter value out of range")

    @rpc_method
    def get_idn(self) -> QMI_InstrumentIdentification:
        """Read instrument type and version and return QMI_InstrumentIdentification instance."""
        self._check_is_open()
        resp = self._ask("*IDN?")
        words = resp.rstrip().split(",")
        if len(words) != 4:
            raise QMI_InstrumentException(f"Unexpected response to *IDN?, got {resp!r}")
        return QMI_InstrumentIdentification(vendor=words[0].strip(),
                                            model=words[1].strip(),
                                            serial=words[2].strip(),
                                            version=words[3].strip())

    @rpc_method
    def reset(self) -> None:
        """Reset the instrument, returning most settings to defaults."""
        self._check_is_open()
        self._cmd("*RST")

    @rpc_method
    def get_num_channels(self) -> int:
        """Return the number of output channels of the instrument."""
        self._check_is_open()
        resp = self._ask(":INST:CAT?")
        channel_names = resp.split(",")
        if channel_names[0].strip().upper() != "T0":
            _logger.error("[%s] Got unexpected list of channels: %s", self._name, resp)
            raise QMI_InstrumentException(f"Instrument reports unexpected channel list: {resp}")
        return len(channel_names) - 1

    @rpc_method
    def get_refclk_source(self) -> RefClkSource:
        """Return the current reference clock source."""
        self._check_is_open()
        resp = self._ask(":PULSE0:ICL:MOD?")
        resp = resp.rstrip("?")  # The instrument returns "XPL?" when the PLL is not locked.
        try:
            return RefClkSource(resp.upper())
        except ValueError:
            # pylint: disable=raise-missing-from
            raise QMI_InstrumentException(f"Unsupported reference clock source {resp!r}")

    @rpc_method
    def is_refclk_pll_locked(self) -> bool:
        """Return True if the reference clock PLL is locked to the external input signal."""
        self._check_is_open()
        resp = self._ask(":PULSE0:ICL:MOD?")
        # The instrument returns "XPL" when the PLL is locked, "XPL?" when the PLL is not locked.
        return (resp.upper() == RefClkSource.EXTPLL.value)

    @rpc_method
    def set_refclk_source(self, source: RefClkSource) -> None:
        """Change the reference clock source.

        Parameters:
            source: New reference clock source (one of the `RefClkSource` constants).
        """
        self._check_is_open()
        source_str = RefClkSource(source).value
        self._cmd(f":PULSE0:ICL:MOD {source_str}")

    @rpc_method
    def get_refclk_rate(self) -> int:
        """Return the expected frequency of the external reference clock in MHz."""
        self._check_is_open()
        resp = self._ask(":PULSE0:ICL:RAT?")
        return self._parse_int(resp)

    @rpc_method
    def set_refclk_rate(self, rate: int) -> None:
        """Change the expected frequency of the external reference clock.

        This parameter is only applicable when an external reference clock source is selected.

        Parameters:
            rate: Expected frequency of the input signal in MHz
                  as an integer (valid range 10 to 100 MHz).
        """
        self._check_range(rate, 10, 100)
        self._check_is_open()
        self._cmd(f":PULSE0:ICL:RAT {rate}")

    @rpc_method
    def get_refclk_level(self) -> float:
        """Return the threshold level for the external clock input in Volt."""
        self._check_is_open()
        resp = self._ask(":PULSE0:ICL:LEV?")
        return self._parse_float(resp)

    @rpc_method
    def set_refclk_level(self, level: float) -> None:
        """Change the threshold level for the external clock input.

        The threshold should be set to 50% of the input signal amplitude.
        This level is only applicable when an external reference clock source is selected.

        Parameters:
            level: Threshold level in Volt (valid range 0.02 to 2.5 Volt).
        """
        self._check_range(level, 0.02, 2.5)
        self._check_is_open()
        self._cmd(f":PULSE0:ICL:LEV {level:.3f}")

    @rpc_method
    def set_refclk_external(self, rate: int, level: float, use_pll: bool) -> None:
        """Configure the instrument to use an external reference clock input.

        Parameters:
            rate:       External clock frequency in MHz (valid range 10 to 100).
            level:      Input threshold level in Volt; should be set to 50% of the signal amplitude
                        (valid range 0.02 to 2.5).
            use_pll:    True to use a PLL to lock to the external input signal.
        """
        source = RefClkSource.EXTPLL if use_pll else RefClkSource.EXTERNAL
        self.set_refclk_source(source)
        self.set_refclk_rate(rate)
        self.set_refclk_level(level)

    @rpc_method
    def get_output_enabled(self) -> bool:
        """Return the global output enable state that applies to all channels."""
        self._check_is_open()
        resp = self._ask(":PULSE0:STAT?")
        return self._parse_bool(resp)

    @rpc_method
    def set_output_enabled(self, enable: bool) -> None:
        """Enable or disable pulse generation.

        This controls a global on/off switch that starts or stops pulse generation
        on all enabled channels. Each channel also has a separate on/off switch
        which only affects that specific channel.

        Parameters:
            enable: True to enable pulse generation (equivalent to pressing the RUN button).
                    False to disable pulse generation (equivalent to pressing the STOP button).
        """
        self._check_is_open()
        arg = 1 if enable else 0
        self._cmd(f":PULSE0:STAT {arg}")

    @rpc_method
    def get_t0_period(self) -> float:
        """Return the T0 pulse interval in seconds."""
        self._check_is_open()
        resp = self._ask(":PULSE0:PER?")
        return self._parse_float(resp)

    @rpc_method
    def set_t0_period(self, period: float) -> None:
        """Change the T0 pulse interval.

        Parameters:
            period: New pulse interval in seconds (valid range 50.0e-9 to 5000.0).
        """
        # Warning: Instrument freaks out when the period value has too many significant digits.
        # Sending fixed point format with 9 digits following the decimal point seems safe.
        self._check_range(period, 50.0e-9, 5000.0)
        self._check_is_open()
        self._cmd(f":PULSE0:PER {period:.9f}")

    @rpc_method
    def get_t0_mode(self) -> PulseMode:
        """Return the T0 pulse generation mode."""
        self._check_is_open()
        resp = self._ask(":PULSE0:MOD?")
        try:
            return PulseMode(resp.upper())
        except ValueError:
            # pylint: disable=raise-missing-from
            raise QMI_InstrumentException(f"Unsupported pulse mode {resp!r}")

    @rpc_method
    def set_t0_mode(self, mode: PulseMode) -> None:
        """Change the T0 pulse generation mode.

        Parameters:
            mode: New pulse generation mode (one of the `PulseMode` constants).
        """
        self._check_is_open()
        mode_str = PulseMode(mode).value
        self._cmd(f":PULSE0:MOD {mode_str}")

    @rpc_method
    def get_t0_burst_count(self) -> int:
        """Return the number of T0 pulses when operating in burst mode."""
        self._check_is_open()
        resp = self._ask(":PULSE0:BCO?")
        return self._parse_int(resp)

    @rpc_method
    def set_t0_burst_count(self, count: int) -> None:
        """Set the number of T0 pulses to generate when operating in burst mode.

        Parameters:
            count: Number of T0 pulses to generate after the output is enabled.
                   Valid range is 1 to 10**6.
        """
        self._check_range(count, 1, 10**6)
        self._check_is_open()
        self._cmd(f":PULSE0:BCO {count}")

    @rpc_method
    def get_t0_duty_cycle(self) -> Tuple[int, int]:
        """Return the parameters for T0 duty cycle mode.

        Returns:
            Tuple (num_pulse, num_skip).
        """
        self._check_is_open()
        resp = self._ask(":PULSE0:PCO?")
        num_pulse = self._parse_int(resp)
        resp = self._ask(":PULSE0:OCO?")
        num_skip = self._parse_int(resp)
        return (num_pulse, num_skip)

    @rpc_method
    def set_t0_duty_cycle(self, num_pulse: int, num_skip: int) -> None:
        """Change the parameters for T0 duty cycle mode.

        Parameters:
            num_pulse: Number of pulses to generate (valid range 1 to 10**6).
            num_skip: Number of pulses to skip (valid range 1 to 10**6).
        """
        self._check_range(num_pulse, 1, 10**6)
        self._check_range(num_skip, 1, 10**6)
        self._check_is_open()
        self._cmd(f":PULSE0:PCO {num_pulse}")
        self._cmd(f":PULSE0:OCO {num_skip}")

    @rpc_method
    def get_trigger_mode(self) -> TriggerMode:
        """Return the T0 trigger mode."""
        self._check_is_open()
        resp = self._ask(":PULSE0:TRIG:MOD?")
        try:
            return TriggerMode(resp.upper())
        except ValueError:
            # pylint: disable=raise-missing-from
            raise QMI_InstrumentException(f"Unsupported trigger mode {resp!r}")

    @rpc_method
    def set_trigger_mode(self, mode: TriggerMode) -> None:
        """Change the T0 trigger mode.

        Parameters:
            mode: New trigger mode (one of the `TriggerMode` constants).
        """
        self._check_is_open()
        mode_str = TriggerMode(mode).value
        self._cmd(f":PULSE0:TRIG:MOD {mode_str}")

    @rpc_method
    def get_trigger_edge(self) -> TriggerEdge:
        """Return the active edge for the T0 trigger."""
        self._check_is_open()
        resp = self._ask(":PULSE0:TRIG:EDGE?")
        try:
            return TriggerEdge(resp.upper())
        except ValueError:
            # pylint: disable=raise-missing-from
            raise QMI_InstrumentException(f"Unsupported trigger edge {resp!r}")

    @rpc_method
    def set_trigger_edge(self, edge: TriggerEdge) -> None:
        """Change the active edge for the T0 trigger.

        Parameters:
            mode: New active trigger edge (one of the `TriggerEdge` constants).
        """
        self._check_is_open()
        edge_str = TriggerEdge(edge).value
        self._cmd(f":PULSE0:TRIG:EDGE {edge_str}")

    @rpc_method
    def get_trigger_level(self) -> float:
        """Return the T0 trigger level in Volt."""
        self._check_is_open()
        resp = self._ask(":PULSE0:TRIG:LEVEL?")
        return self._parse_float(resp)

    @rpc_method
    def set_trigger_level(self, level: float) -> None:
        """Change the T0 trigger level.

        Parameters:
            level: New trigger level in Volt (valid range 0.2 .. 15).
        """
        self._check_range(level, 0.2, 15.0)
        self._check_is_open()
        self._cmd(f":PULSE0:TRIG:LEVEL {level:.3f}")

    @rpc_method
    def get_channel_enabled(self, channel: int) -> bool:
        """Return the output enable state for the specified channel.

        Parameters:
            channel: Channel index in range 1 .. `num_channels`.
        """
        if channel < 1:
            raise ValueError("Invalid channel index")
        self._check_is_open()
        resp = self._ask(f":PULSE{channel}:STAT?")
        return self._parse_bool(resp)

    @rpc_method
    def set_channel_enabled(self, channel: int, enable: bool) -> None:
        """Enable or disable output for the specified channel.

        Parameters:
            channel: Channel index in range 1 .. `num_channels`.
            enable:  `True` to enable the channel, `False` to disable.
        """
        self._check_range(channel, 1, None)
        self._check_is_open()
        arg = 1 if enable else 0
        self._cmd(f":PULSE{channel}:STAT {arg}")

    @rpc_method
    def get_channel_width(self, channel: int) -> float:
        """Return the configured pulse width for the specified channel.

        Parameters:
            channel: Channel index in range 1 .. `num_channels`.

        Returns:
            Configured pulse width in seconds.
        """
        self._check_range(channel, 1, None)
        self._check_is_open()
        resp = self._ask(f":PULSE{channel}:WIDT?")
        return self._parse_float(resp)

    @rpc_method
    def set_channel_width(self, channel: int, width: float) -> None:
        """Configure pulse width for the specified channel.

        Parameters:
            channel: Channel index in range 1 .. `num_channels`.
            width:   New pulse width in seconds (valid range 10e-9 .. 999.9).
        """
        self._check_range(channel, 1, None)
        self._check_range(width, 10.0e-9, 999.9)
        self._check_is_open()
        self._cmd(f":PULSE{channel}:WIDT {width:.11f}")

    @rpc_method
    def get_channel_delay(self, channel: int) -> float:
        """Return the configured delay for the specified channel.

        Parameters:
            channel: Channel index in range 1 .. `num_channels`.

        Returns:
            Configured delay in seconds.
        """
        self._check_range(channel, 1, None)
        self._check_is_open()
        resp = self._ask(f":PULSE{channel}:DEL?")
        return self._parse_float(resp)

    @rpc_method
    def set_channel_delay(self, channel: int, delay: float) -> None:
        """Configure delay for the specified channel.

        Parameters:
            channel: Channel index in range 1 .. `num_channels`.
            delay:   New delay in seconds (valid range -99.9 .. 999.9).
        """
        self._check_range(channel, 1, None)
        self._check_range(delay, -99.9, 999.9)
        self._check_is_open()
        self._cmd(f":PULSE{channel}:DEL {delay:.11f}")

    @rpc_method
    def get_channel_mode(self, channel: int) -> PulseMode:
        """Return the pulse generation mode of the specified channel."""
        self._check_range(channel, 1, None)
        self._check_is_open()
        resp = self._ask(f":PULSE{channel}:CMOD?")
        try:
            return PulseMode(resp.upper())
        except ValueError:
            # pylint: disable=raise-missing-from
            raise QMI_InstrumentException(f"Unsupported pulse mode {resp!r}")

    @rpc_method
    def set_channel_mode(self, channel: int, mode: PulseMode) -> None:
        """Change the pulse generation mode of the specified channel.

        Parameters:
            channel: Channel index in range 1 .. `num_channels`.
            mode:    New pulse mode (one of the `PulseMode` constants).
        """
        self._check_range(channel, 1, None)
        self._check_is_open()
        mode_str = PulseMode(mode).value
        self._cmd(f":PULSE{channel}:CMOD {mode_str}")

    @rpc_method
    def get_channel_burst_count(self, channel: int) -> int:
        """Return the number of pulses to output when operating in burst mode."""
        self._check_range(channel, 1, None)
        self._check_is_open()
        resp = self._ask(f":PULSE{channel}:BCO?")
        return self._parse_int(resp)

    @rpc_method
    def set_channel_burst_count(self, channel: int, count: int) -> None:
        """Set the number of T0 pulses to generate when operating in burst mode.

        These parameters only has effect when the channel is operating in `PulseMode.BURST` mode.

        Parameters:
            channel:    Channel index in range 1 .. `num_channels`.
            count:      Number of pulses to generate after the output is enabled.
                        Valid range is 1 to 10**6.
        """
        self._check_range(channel, 1, None)
        self._check_range(count, 1, 10**6)
        self._check_is_open()
        self._cmd(f":PULSE{channel}:BCO {count}")

    @rpc_method
    def get_channel_duty_cycle(self, channel: int) -> Tuple[int, int]:
        """Return the duty cycel parameters for the specified channel.

        Returns:
            Tuple (num_pulse, num_skip).
        """
        self._check_range(channel, 1, None)
        self._check_is_open()
        resp = self._ask(f":PULSE{channel}:PCO?")
        num_pulse = self._parse_int(resp)
        resp = self._ask(f":PULSE{channel}:OCO?")
        num_skip = self._parse_int(resp)
        return (num_pulse, num_skip)

    @rpc_method
    def set_channel_duty_cycle(self, channel: int, num_pulse: int, num_skip: int) -> None:
        """Change the duty cycle parameters for the specified channel.

        These parameters only have effect when the channel is operating in `PulseMode.DUTYCYCLE` mode.

        Parameters:
            channel:    Channel index in range 1 .. `num_channels`.
            num_pulse:  Number of pulses to generate (valid range 1 to 10**6).
            num_skip:   Number of pulses to skip (valid range 1 to 10**6).
        """
        self._check_range(channel, 1, None)
        self._check_range(num_pulse, 1, 10**6)
        self._check_range(num_skip, 1, 10**6)
        self._check_is_open()
        self._cmd(f":PULSE{channel}:PCO {num_pulse}")
        self._cmd(f":PULSE{channel}:OCO {num_skip}")

    @rpc_method
    def get_output_driver(self, channel: int) -> OutputDriver:
        """Return the current output mode for the specified channel."""
        self._check_range(channel, 1, None)
        self._check_is_open()
        resp = self._ask(f":PULSE{channel}:OUTP:MODE?")
        try:
            return OutputDriver(resp.upper())
        except ValueError:
            # pylint: disable=raise-missing-from
            raise QMI_InstrumentException(f"Unsupported output mode {resp!r}")

    @rpc_method
    def set_output_driver(self, channel: int, mode: OutputDriver) -> None:
        """Change the output mode for the specified channel.

        Parameters:
            channel: Channel index in range 1 .. `num_channels`.
            mode:    New output mode (one of the `OutputDriver` constants).
        """
        self._check_range(channel, 1, None)
        self._check_is_open()
        mode_str = OutputDriver(mode).value
        self._cmd(f":PULSE{channel}:OUTP:MODE {mode_str}")

    @rpc_method
    def get_output_amplitude(self, channel: int) -> float:
        """Return the current output amplitude in Volt.

        The returned value only applies when the output mode is `OutputMode.ADJUSTABLE`.
        """
        self._check_range(channel, 1, None)
        self._check_is_open()
        resp = self._ask(f":PULSE{channel}:OUTP:AMPL?")
        return self._parse_float(resp)

    @rpc_method
    def set_output_amplitude(self, channel: int, amplitude: float) -> None:
        """Change the output amplitude of the specified channel..

        The specified amplitude represents the open-circuit voltage of the output.
        The output channels have 50 Ohm output impedance. As a result, the voltage
        levels will drop to 50% of the specified amplitude when a 50 Ohm load
        is connected to the output port.

        This setting only applies when the output mode is `OutputMode.ADJUSTABLE`.

        Parameters:
            channel:    Channel index in range 1 .. `num_channels`.
            amplitude:  New output-high level in Volt (valid range 2.0 to 20.0).
        """
        self._check_range(channel, 1, None)
        self._check_range(amplitude, 2.0, 20.0)
        self._check_is_open()
        self._cmd(f":PULSE{channel}:OUTP:AMPL {amplitude:.3f}")

    @rpc_method
    def get_output_inverted(self, channel: int) -> bool:
        """Return True if the specified channel is active low."""
        self._check_range(channel, 1, None)
        self._check_is_open()
        resp = self._ask(f":PULSE{channel}:POL?")
        return resp.upper() != "NORM"

    @rpc_method
    def set_output_inverted(self, channel: int, invert: bool) -> None:
        """Change the output polarity of the specified channel.

        Parameters:
            channel: Channel index in range 1 .. `num_channels`.
            invert:  True to select active low output, False to select active high.
        """
        self._check_range(channel, 1, None)
        self._check_is_open()
        arg = "INV" if invert else "NORM"
        self._cmd(f":PULSE{channel}:POL {arg}")

    @rpc_method
    def set_display_enabled(self, enable: bool) -> None:
        """Enable or disable the display on the front panel of the instrument.

        Disabling the display also blocks the front panel keys, except for
        the power button which always works.

        Parameters:
            enable: True to enable display and keypad, False to disable.
        """
        self._check_is_open()
        arg = 1 if enable else 0
        self._cmd(f":DISP:ENAB {arg}")



"""Implementation of class KeyboardReader.

The KeyboardReader provides a way to read lines from the keyboard without blocking
in a platform-independent way.
"""

import sys

# We implement a platform-specific _PollKeyboard class that provides a 'poll_line' method to
# fetch a single line of input in a non-blocking way.
#
# The reason that we implement this as a class is that in Windows, the 'msvcrt' API doesn't
# do buffering for us, so we have to buffer the characters ourselves.
#
# To allow identical interfaces between Windows and Unix platforms, we begrudgingly implement the same
# class-based interface in Unix, even though in Unix, buffering is handled by the OS.

if sys.platform.startswith("win"):

    # _PollKeyboard class for the Windows platform.

    import msvcrt  # Only exists in Windows.

    class _PollKeyboard:
        """Internal helper class to read from the keyboard on Windows platforms."""

        def __init__(self) -> None:
            self._buffer = ""

        def poll_line(self) -> str:
            """Poll for a single line of input on the Windows platform.

            Returns:
                An empty string if no Enter-terminated line is available;
                a non-empty string containing a line and ending in '\n' if an Enter-terminated string is available.
            """

            # We read up to 256 characters using msvcrt's getwch() method, as long as input is available.
            #
            # In Windows, the Enter key produces an end-of-line character ('\r').
            # If this character is detected, we the buffer followed by a newline character ('\n') instead.

            i = 0
            while (i < 256) and msvcrt.kbhit():
                i += 1
                c = msvcrt.getwch()
                if c == '\r':
                    # We replace the carriage return character ('\r') by a linefeed character ('\n').
                    ret = self._buffer + '\n'
                    self._buffer = ""  # Clear the buffer
                    return ret
                else:
                    self._buffer += c
            return ""

else:
    # _PollKeyboard class for Unix platforms (including macOS which is a BSD derivative).

    import select

    class _PollKeyboard:
        """Internal helper class to read from the keyboard on Unix platforms."""

        def poll_line(self) -> str:
            """Poll for a single line of input on a Unix platform.

            Returns:
                An empty string if no Enter-terminated line is available;
                a non-empty string containing a line and ending in '\n' if an Enter-terminated string is available.
            """

            # Use select() to determine if a line of input is available.
            #
            # The OS does the line buffering for us. Hitting 'Enter' on the keyboard will insert a
            # newline character ('\n') into the buffer, and will make the selector generate an event.
            #
            # We use the "select" module instead of the "selectors" module
            # because the "selectors" module raises a PermissionError
            # when stdin has been redirected from /dev/null. This happens
            # with server processes when they run in the background.

            (rl, wl, xl) = select.select([sys.stdin], [], [], 0)
            if rl:
                return sys.stdin.readline()
            else:
                return ""


class KeyboardReader:
    """A class that implements non-blocking read of single lines from the keyboard."""

    def __init__(self) -> None:
        self._poll_keyboard = _PollKeyboard()  # Instantiate platform-dependent keyboard line poller.

    def poll_line(self) -> str:
        """Read a line from the keyboard without blocking.

        If keyboard input is waiting, return a string containing one line of input.
        Otherwise, return an empty string.
        """
        return self._poll_keyboard.poll_line()

    def clear_buffer(self) -> None:
        """Remove pending data from the keyboard input buffer."""
        while self.poll_line():
            pass

    def poll_quit(self) -> bool:
        """Check if the user has typed "Q" to stop the measurement.

        Return True if the user has typed "Q", otherwise return False.
        """
        line = self.poll_line()
        if line:
            if line.strip().upper() == "Q":
                return True
            else:
                print("Measurement in progress - type 'Q' followed by Enter to quit.")
        return False



#! /usr/bin/env python3

import time

import qmi
from qmi.instruments.cobolt.laser_06_01 import Cobolt_Laser_06_01
from qmi.instruments.thorlabs.pm100d import Thorlabs_PM100D

laser_transport = "serial:/dev/ttyACM0"
pm_device = "USB0::4883::32888::P0019304::0::INSTR"

qmi.start("calibration_server", "qmi.conf")

laser = qmi.make_instrument("laser", Cobolt_Laser_06_01, laser_transport)
pm    = qmi.make_instrument("pm", Thorlabs_PM100D, pm_device, '@py')

laser.open()
pm.open()

time.sleep(0.100)
input("\nRelay server active, press Enter to quit ...\n")




""" QMI instrument driver for the PicoQuant HydraHarp 400 instrument.

The instrument driver makes use of the manufacturer provided software libraries, "hhlib.so" for Linux OS,
or "hhlib.dll" or "hhlib64.dll" for 32-bit and 64-bit Windows OS, respectively.
Please find the licence terms for these files in the dedicated software package for the HydraHarp instrument at
https://www.picoquant.com/products/category/tcspc-and-time-tagging-modules/hydraharp-400-multichannel-picosecond-event-timer-tcspc-module
--> "Software" tab --> download link in "Current software and developer's library version".
"""
import ctypes
import enum
import logging
from typing import List, Tuple

from qmi.core.exceptions import QMI_InvalidOperationException
from qmi.core.rpc import rpc_method
from qmi.instruments.picoquant.support._library_wrapper import _LibWrapper
from qmi.instruments.picoquant._picoquant import _str_to_enum, _PicoquantHarp, _EDGE
from qmi.instruments.picoquant.support._events import _MODE

_logger = logging.getLogger(__name__)


@enum.unique
class _REFSRC(enum.Enum):
    """Symbolic constants for the :func:`~HydraHarpDevice.initialize` method's `refsource` argument.

    These are defined as preprocessor symbols in the ``hhdefin.h`` C header file.
    """
    INTERNAL = 0
    """Use internal clock."""

    EXTERNAL = 1
    """Use external clock."""


@enum.unique
class _MEASCTL(enum.Enum):
    """Symbolic constants for the :func:`~HydraHarpDevice.setMeasurementControl` `control` argument.

    These are defined as preprocessor symbols in the ``hhdefin.h`` C header file.
    """
    SINGLESHOT_CTC = 0
    """Default value. Acquisition starts by software command and runs until CTC expires. The duration
    is set by the tacq parameter passed to HH_StartMeas."""
    C1_GATED = 1
    """Histograms are collected for the period where C1 is active. This can be the logical high or
    low period dependent on the value supplied to the parameter startedge."""
    C1_START_CTC_STOP = 2
    """Data collection is started by a transition on C1 and stopped by expiration of the internal CTC. 
    Which transition actually triggers the start is given by the value supplied to the parameter startedge. 
    The duration is set by the tacq parameter passed to HH_StartMeas."""
    C1_START_C2_STOP = 3
    """Data collection is started by a transition on C1 and stopped by by a transition on C2. Which transitions
    actually trigger start and stop is given by the values supplied to the parameters startedge and stopedge."""
    CONT_C1_GATED = 4
    """Histograms are collected for each period where C1 is active. This can be the logical high or low periods
    dependent on the value supplied to the parameter startedge."""
    CONT_C1_START_CTC_STOP = 5
    """Histogram collection is started by a transition on C1 and stopped by expiration of the internal CTC.
    Which transition actually triggers the start is given by the value supplied to the parameter startedge.
    Histogram duration is set by the tacq parameter passed to HH_StartMeas. The current histogram
    ends if a new trigger occurs before the CTC has expired."""
    CONT_CTC_RESTART = 6
    """Histogram collection is started and stopped exclusively by the internal CTC. Consecutive histograms will
    line up without gaps. Histogram duration is set by the tacq parameter passed to HH_StartMeas."""


@enum.unique
class _FLAG(enum.IntFlag):
    """Bitfield constants for the return value of the :func:`~HydraHarpDevice.getFlags` function.

    These are defined as preprocessor symbols in the ``hhdefin.h`` C header file.
    """
    OVERFLOW = 0x0001
    """Histogram mode only. It indicates that a histogram measurement has reached the maximum count 
    as specified via HH_SetStopOverflow."""
    FIFOFULL = 0x0002
    """TTTR mode only.  It indicates that the data FiFo has run full. The measurement will then have
    to be aborted as data integrity is no longer maintained."""
    SYNC_LOST = 0x0004
    """This flag may occur in T3 mode and in histo mode. It indicates that the sync signal has been lost
    which in this case is critical as the function of T3 mode and histo mode relies on an uninterrupted
    sync signal."""
    REF_LOST = 0x0008
    """This flag will occur when the HydraHarp is programmed to use an external reference clock and this
    reference clock is lost."""
    SYSERROR = 0x0010
    """Indicates an error of the hardware or internal software. The user should in this case
    call the library routine HH_GetHardwareDebugInfo and provide the result to PicoQuant support."""
    ACTIVE = 0x0020
    """Measurement is running."""
    CNTS_DROPPED = 0x0040
    """Indicates that counts were dropped at the first level FiFo directly following the TDC of an input channel.
    This occurs typically only at extremely high count rates. Dependent on the application scenario this may or
    may not be considered critical."""


@enum.unique
class _WARNING(enum.IntFlag):
    """Bitfield constants for the return value of the :func:`~HydraHarpDevice.getWarnings` function.

    These are defined as preprocessor symbols in the ``hhdefin.h`` C header file. For full descriptions
    see the chapter '8.1. Warnings' from the HydraHarp HHLib Manual.
    """
    SYNC_RATE_ZERO = 0x0001
    """No pulses are detected at the sync input. In histogramming and T3 mode this is crucial and the
    measurement will not work without this signal."""
    SYNC_RATE_TOO_LOW = 0x0002
    """The detected pulse rate at the sync input is below 100 Hz and cannot be determined accurately.
    Other warnings may not be reliable under this condition."""
    SYNC_RATE_TOO_HIGH = 0x0004
    """The pulse rate at the sync input (after the divider) is higher than 75 MHz. This is close to 
    the TDC limit. Sync events will be lost above 78 MHz."""
    INPT_RATE_ZERO = 0x0010
    """No counts are detected at any of the input channels."""
    INPT_RATE_TOO_HIGH = 0x0040
    """The overall pulse rate at the input channels is higher than 80 MHz (USB 3.0 connection) or higher
    than 9 MHz (USB 2.0 connection). The measurement will likely lead to a FIFO overrun."""
    INPT_RATE_RATIO = 0x0100
    """This warning is issued in histogramming and T3 mode when the rate at any input channel is higher
    than 5% of the sync rate."""
    DIVIDER_GREATER_ONE = 0x0200
    """In T2 mode: The sync divider is set larger than 1. This is probably not intended.
    In histogramming and T3 mode: If the pulse rate at the sync input is below 75 MHz then a divider
    >1 is not needed."""
    TIME_SPAN_TOO_SMALL = 0x0400
    """This warning is issued in histogramming and T3 mode when the sync period (1/SyncRate) is longer
    than the start to stop time span that can be covered by the histogram or by the T3 mode records."""
    OFFSET_UNNECESSARY = 0x0800
    """This warning is issued in histogramming and T3 mode when an offset >0 is set even though the sync
    period (1/SyncRate) can be covered by the measurement time span without using an offset."""


class PicoQuant_HydraHarp400(_PicoquantHarp):
    """Instrument driver for the PicoQuant HydraHarp 400."""

    _MODEL = "HH"
    _MAXDEVNUM = 8
    _TTREADMAX = 131072

    @property
    def _max_dev_num(self):
        return self._MAXDEVNUM

    @property
    def _ttreadmax(self):
        return self._TTREADMAX

    @property
    def _model(self):
        return self._MODEL

    @property
    def _lib(self) -> _LibWrapper:
        if self._lazy_lib is None:  # type: ignore
            self._lazy_lib = _LibWrapper('HH')
        return self._lazy_lib

    @rpc_method
    def initialize(self, mode_str: str, refsource_str: str) -> None:
        """Initialize the device.

        This routine must be called before any of the other methods, except OpenDevice, CloseDevice,
        GetErrorString and GetLibraryVersion can be used.

        Arguments:
            mode_str (str): Opening mode. Can be any of 'HIST', 'T2', 'T3', 'CONT'.
                            The latest driver version V3.0+ supports T2, T3 and CONT modes.
            refsource_str (str): Reference source for time. Can be any of 'INTERNAL', 'EXTERNAL'.

        Raises:
            ValueError: if either argument is an unknown string value.
            QMI_InstrumentException: in case of a library error.
        """
        self._check_is_open()

        if self._measurement_running:
            raise QMI_InvalidOperationException("Measurement still active")
        mode = _str_to_enum(_MODE, mode_str)
        if mode not in (_MODE.HIST, _MODE.T2, _MODE.T3, _MODE.CONT):
            raise ValueError("Unsupported measurement mode")

        if self._lib_version.startswith("2") and mode in (_MODE.T3, _MODE.CONT):
            raise ValueError(f"Unsupported measurement in mode library version {self._lib_version}")

        refsource = _str_to_enum(_REFSRC, refsource_str)

        with self._device_lock:
            self._lib.Initialize(self._devidx, mode.value, refsource.value)
            self._mode = mode

        # Reset the event filter configuration.
        self.set_event_filter(reset_filter=True)
        self.set_block_events(False)
        _logger.info("[%s] Initialized device with mode %s and source %s", self._name, mode.value, refsource.value)

    @rpc_method
    def get_module_info(self) -> List[Tuple[int, int]]:
        self._check_is_open()
        info = []
        with self._device_lock:
            nummod = ctypes.c_int()
            self._lib.GetNumOfModules(self._devidx, nummod)
            for modidx in range(nummod.value):
                modelcode = ctypes.c_int()
                versioncode = ctypes.c_int()
                self._lib.GetModuleInfo(self._devidx, modidx, modelcode, versioncode)
                info.append((modelcode.value, versioncode.value))
        return info

    @rpc_method
    def set_sync_channel_offset(self, value: int) -> None:
        self._check_is_open()
        with self._device_lock:
            self._lib.SetSyncChannelOffset(self._devidx, value)

    @rpc_method
    def set_measurement_control(self, meascontrol_str: str, startedge_str: str, stopedge_str: str) -> None:
        self._check_is_open()
        meascontrol = _str_to_enum(_MEASCTL, meascontrol_str)
        startedge = _str_to_enum(_EDGE, startedge_str)
        stopedge = _str_to_enum(_EDGE, stopedge_str)
        with self._device_lock:
            self._lib.SetMeasControl(self._devidx, meascontrol.value, startedge.value, stopedge.value)

    @rpc_method
    def get_flags(self) -> List[str]:
        self._check_is_open()
        with self._device_lock:
            bitset = ctypes.c_int()
            self._lib.GetFlags(self._devidx, bitset)
            flags = _FLAG(bitset.value)
            return [flag.name for flag in _FLAG if flag in flags and flag.name is not None]

    @rpc_method
    def get_warnings(self) -> List[str]:
        self._check_is_open()
        with self._device_lock:
            bitset = ctypes.c_int()
            self._lib.GetWarnings(self._devidx, bitset)
            warnings = _WARNING(bitset.value)
            return [warning.name for warning in _WARNING if warning in warnings and warning.name is not None]

    @rpc_method
    def calibrate(self) -> None:
        self._check_is_open()
        with self._device_lock:
            self._lib.Calibrate(self._devidx)

    @rpc_method
    def set_sync_cfd(self, level: int, zc: int) -> None:
        self._check_is_open()
        with self._device_lock:
            self._lib.SetSyncCFD(self._devidx, level, zc)

    @rpc_method
    def set_input_cfd(self, channel: int, level: int, zc: int) -> None:
        self._check_is_open()
        with self._device_lock:
            self._lib.SetInputCFD(self._devidx, channel, level, zc)
